"""Deal with the comment sections in a component definition."""
from typing import TYPE_CHECKING, List

from raesl.compile import diagnostics
from raesl.compile.ast import comment_storage
from raesl.compile.ast.components import get_doc_comment_comp_elements

if TYPE_CHECKING:
    from raesl.compile.ast.comment_storage import DocCommentDistributor
    from raesl.compile.ast.components import ComponentDefinition
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CompDefCommentBuilder:
    """Collect the names in the 'comments' section, and hook them into the doc comments
    distributor.

    Arguments:
        comp_child_builders: Component definition's section builders storage.

    Attributes:
        diag_store: Diagnostics storage of component definition child builders.
        name_toks: Names occurring in a comment section, collected during parsing.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.name_toks: List["Token"] = []

    def add_comment(self, name_tok: "Token"):
        """Parser found a name in a comments section, store it for future processing."""
        self.name_toks.append(name_tok)

    def finish_comp(
        self, comp_def: "ComponentDefinition", doc_distributor: "DocCommentDistributor"
    ):
        """Process all collected names. This method should be the final step in
        processing a component definition, as it needs all elements that take doc
        comments.

        Arguments:
            comp_def: Component definition to finish.
            doc_distributor: Object that distributes doc comments to interested elements
                of the specification.
        """
        comp_def_doc_elements = get_doc_comment_comp_elements(comp_def)
        available = {}  # Available elements in the specification, ordered by their name.
        for elm in comp_def_doc_elements:
            assert elm.doc_tok is not None
            available[elm.doc_tok.tok_text] = elm

        for name_tok in self.name_toks:
            # Find language elements to point to by their main name only.
            i = name_tok.tok_text.find(".")
            if i < 0:
                main_name = name_tok.tok_text
            else:
                main_name = name_tok.tok_text[:i]

            opt_elm = available.get(main_name)
            if opt_elm is None:
                # Report an error if an element is not available.
                if comp_def.name_tok is None:
                    comp_name = "world"
                else:
                    comp_name = comp_def.name_tok.tok_text

                # Can't find doc element in component.
                self.diag_store.add(
                    diagnostics.E205(
                        f"element '{name_tok.tok_text}'",
                        f"component '{comp_name}'",
                        name_tok.get_location(),
                    )
                )

                # Construct a dummy element so any comment after it is caught.
                # No errors to report as the above already reported one.
                doc_distributor.add_dummy_element(name_tok, False)

            else:
                # Add a proxy, redirecting doc comments to the correct element.
                doc_distributor.add_element(comment_storage.ProxyDocStore(name_tok, opt_elm))



"""Node classes representing elementary and combined flows."""
from typing import TYPE_CHECKING, List, Optional

from raesl.compile.ast import comment_storage
from raesl.utils import split_first_dot

if TYPE_CHECKING:
    from raesl.compile.ast.types import BaseType
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import Counter


class Node:
    """Abstract class for nodes.

    Note that a Node only has a name. The typed sub-tree starts with VarNode.

    Arguments:
        name_tok: Name of the node.
    """

    def __init__(self, name_tok: "Token"):
        self.name_tok = name_tok


class VarNode(Node):
    """Abstract base class of a variable or parameter node that can be shared with
    variable groups and other users such as transformations and goals.

    Arguments:
        name_tok: Token with the name/location of the variable or parameter.
        the_type: Type of the variable or parameter.
    """

    def __init__(self, name_tok: "Token", the_type: "BaseType"):
        super(VarNode, self).__init__(name_tok)
        self.the_type = the_type

        assert name_tok.tok_type == "NAME"  # Should be a plain identifier.

    def resolve_node(self, name: str) -> Optional["VarNode"]:
        """Find the varparam (sub)node that matches the provided dotted 'name'.

        Arguments:
            name: Possibly dotted name that should point at an existing sub-node.
                The empty string denotes 'self'.

        Returns:
            The node that matches the name, or None if no such node exists. In the
                latter case, use 'self.get_error_position(name)' to get an indication
                where the match fails in the name.
        """
        raise NotImplementedError("Implement me in {}.".format(repr(self)))

    def get_error_position(self, name: str) -> int:
        """Return the index in the given string where an error occurs in resolving the
        node.

        Arguments:
            name: Name of the element to find.

        Returns:
            Approximated index in the string where matching the element fails.
                Returned value has no meaning if resolving a node succeeds.
        """
        raise NotImplementedError("Implement me in {}.".format(repr(self)))

    def add_comment(self, comment_tok: "Token"):
        """Add found documentation comment.

        Arguments:
            comment_tok: The raw documentation token to add.
        """
        raise NotImplementedError("Implement me in {}.".format(repr(self)))


class ElementaryVarNode(VarNode, comment_storage.DocElement):
    """Elementary variable/parameter node.

    Arguments:
        name_tok: Token with the name/position of the variable or parameter.
        the_type: Type of the variable or parameter.
        counter: Object to give out unique identification numbers, yet be resistant
            against re-use of imported modules.

    Attributes:
        id: Unique number of the node, mostly useful for dumps and debugging.
        comments: Stored comments of the node.
    """

    def __init__(self, name_tok: "Token", the_type: "BaseType", counter: "Counter"):
        super(ElementaryVarNode, self).__init__(name_tok, the_type)
        self.id = counter.next()
        self.comments: List[str] = []

    def resolve_node(self, name: str) -> Optional[VarNode]:
        if name != "":
            return None
        return self

    def get_error_position(self, name: str) -> int:
        return 0

    def add_comment(self, comment_tok: "Token"):
        """Add found documentation comment.

        Arguments:
            comment_tok: The raw documentation token to add.
        """
        self.comments.extend(comment_storage.decode_doc_comments(comment_tok))

    def get_comment(self):
        return self.comments

    def __repr__(self):
        return "ElementaryVarNode[" + str(self.id) + "]"


class CompoundVarNode(VarNode, comment_storage.DocAddElement):
    """Grouped variable/parameter node.

    Arguments:
        name_tok: Token with the name/position of the variable or parameter.
        the_type: Type of the variable or parameter.
        child_nodes: Child nodes of the group.

    Attributes:
        name_index: Mapping of name to the associated VarNode instance.
    """

    def __init__(self, name_tok: "Token", the_type: "BaseType", child_nodes: List[VarNode]):
        super(CompoundVarNode, self).__init__(name_tok, the_type)
        self.child_nodes = child_nodes
        self.name_index = dict((cn.name_tok.tok_text, cn) for cn in child_nodes)

        assert isinstance(self.child_nodes, list)

        # Bundle doesn't have duplicate child names.
        assert len(child_nodes) == len(self.name_index)

    def resolve_node(self, name: str) -> Optional[VarNode]:
        local_name, remaining_name, _dot_length = split_first_dot(name)
        if local_name == "":
            return self

        child_node = self.name_index.get(local_name)
        if child_node is None:
            return None
        return child_node.resolve_node(remaining_name)

    def get_error_position(self, name: str) -> int:
        local_name, remaining_name, dot_length = split_first_dot(name)

        child_node = self.name_index.get(local_name)
        if child_node is None:
            return 0  # Local name == first name is wrong.
        else:
            # Ask child about the position of the error.
            offset = child_node.get_error_position(remaining_name)
            return offset + len(local_name) + dot_length

    def add_comment(self, comment_tok: "Token"):
        """Compound node doesn't own a store, push comment down to all children."""
        for child in self.child_nodes:
            child.add_comment(comment_tok)


class GroupNode(Node):
    """Class describing content of a variable group. Unlike the VarNode above,
    a group node has no type of its own.

    Arguments:
        name_tok: Name of the group node.
        child_nodes: Elements of the group.
    """

    def __init__(self, name_tok: "Token", child_nodes: List[Node]):
        super(GroupNode, self).__init__(name_tok)
        self.child_nodes = child_nodes



"""Line matcher state machines for goals."""
from typing import TYPE_CHECKING

from raesl.compile.ast import components
from raesl.compile.machine_files import sub_clause, typing
from raesl.compile.machine_files.machine_parts import (
    get_argument_references_part,
    get_does_auxiliary_verb_part,
)

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_GOAL_HEADER_SPEC = """
goal_header:
    start initial;
    start -> s1 [GOAL_REQUIREMENT_KW] tag=goal_kind;
    start -> s1 [GOAL_CONSTRAINT_KW] tag=goal_kind;

    end accept=goal_header;
    s1 -> end [NL_TK];
"""


def _process_goal_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    goal_kind = tags["goal_kind"][0]
    builder.notify_new_section(goal_kind, False)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.new_goal_header(goal_kind)


_GOAL_MAIN_WITH_SUBS_SPEC = (
    """
goal_main_with_subs:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
    s2 -> s3 [NAME] tag=active_compname;
"""
    + get_does_auxiliary_verb_part("s3", ["s4"], tagname="doesaux")
    + """\
    s4 -> s5 [NAME] tag=verb;
"""
    + get_argument_references_part("s5", ["s7"], tagname="argument_name")
    + """\
    s7 -> s8 [NAME] tag=prepos;
    s8 -> s9 [NAME] tag=passive_compname;

    s9 -> s10 [WITH_KW];
    s10 -> s11 [SUB_CLAUSES_KW] tag=with_subclause;

    end accept=goal_main_with_subs;
    s11 -> end [NL_TK];
    s11 -> end [EOF_TK];
"""
)


def _process_goal_with_subs(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    label = tags["label"][0]
    active = tags["active_compname"][0]
    doesaux = tags["doesaux"][0]
    verb = tags["verb"][0]
    flows = [components.Flow(name_tok) for name_tok in tags["argument_name"]]
    prepos = tags["prepos"][0]
    passive = tags["passive_compname"][0]
    goal = components.Goal(label, active, doesaux, verb, flows, prepos, passive)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_goal(goal)


_GOAL_MAIN_NO_SUBS_SPEC = (
    """
goal_main_no_subs:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
    s2 -> s3 [NAME] tag=active_compname;
"""
    + get_does_auxiliary_verb_part("s3", ["s4"], tagname="doesaux")
    + """\
    s4 -> s5 [NAME] tag=verb;
"""
    + get_argument_references_part("s5", ["s7"], tagname="argument_name")
    + """\
    s7 -> s8 [NAME] tag=prepos;
    s8 -> s9 [NAME] tag=passive_compname;

    end accept=goal_main_no_subs;
    s9 -> end [NL_TK];
    s9 -> end [EOF_TK];
"""
)


def _process_goal_no_subs(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    label = tags["label"][0]
    active = tags["active_compname"][0]
    doesaux = tags["doesaux"][0]
    verb = tags["verb"][0]
    flows = [components.Flow(name_tok) for name_tok in tags["argument_name"]]
    prepos = tags["prepos"][0]
    passive = tags["passive_compname"][0]
    goal = components.Goal(label, active, doesaux, verb, flows, prepos, passive)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_goal(goal)


def _process_goal_subclause(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    sub = sub_clause.decode_subclause(tags)
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_goal_subclause(sub)


MACHINES: typing.MachineTripletList = [
    ("GOAL_HEADER_MACHINE", _GOAL_HEADER_SPEC, _process_goal_header),
    ("GOAL_MAIN_WITH_SUBS_MACHINE", _GOAL_MAIN_WITH_SUBS_SPEC, _process_goal_with_subs),
    ("GOAL_MAIN_NO_SUBS_MACHINE", _GOAL_MAIN_NO_SUBS_SPEC, _process_goal_no_subs),
    ("GOAL_SUB_CLAUSE_MACHINE", sub_clause.SUB_CLAUSE_SPEC, _process_goal_subclause),
]



"""Builder for collecting and building component definitions."""
import collections
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Union

from raesl.compile import diagnostics
from raesl.compile.ast import components
from raesl.compile.typechecking.compdef_behavior_builder import (
    CompDefBehaviorBuilder,
    ParsedBehavior,
)
from raesl.compile.typechecking.compdef_comment_builder import CompDefCommentBuilder
from raesl.compile.typechecking.compdef_compinst_builder import (
    CompDefCompInstBuilder,
    ComponentInstance,
)
from raesl.compile.typechecking.compdef_design_builder import CompDefDesignBuilder
from raesl.compile.typechecking.compdef_goal_builder import CompDefGoalBuilder
from raesl.compile.typechecking.compdef_need_builder import CompDefNeedBuilder
from raesl.compile.typechecking.compdef_relinst_builder import (
    CompDefRelInstBuilder,
    RelInst,
)
from raesl.compile.typechecking.compdef_transform_builder import CompDefTransformBuilder
from raesl.compile.typechecking.compdef_vargroup_builder import (
    CollectedVarGroup,
    CompDefVarGroupBuilder,
)
from raesl.compile.typechecking.compdef_varparam_builder import CompDefVarParamBuilder
from raesl.compile.typechecking.orderer import Orderer

if TYPE_CHECKING:
    from raesl.compile.ast import exprs
    from raesl.compile.ast.comment_storage import DocCommentDistributor
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder


class Counter:
    """Class for handing out unique numeric values.

    Normally, a static class variable would do, except testing more than one
    specification at a time doesn't reset the counter, leading to different output
    depending on what is being tested together.

    Arguments:
        first_free_value: First free value to set the counter to.

    Attributes:
        counter: Next free value.
    """

    def __init__(self, first_free_value: int):
        self.counter = first_free_value

    def next(self) -> int:
        """Get a unique number from the counter instance."""
        value = self.counter
        self.counter = self.counter + 1
        return value


class CompDefChildBuilders:
    """Class storing child builders for all sections of a component definition.

    As type checking cannot be done until the entire specification has been parsed
    (global types, verbs, and relation definitions may not exist at the time of the end
    of a component definition, and definitions of instantiated component may be defined
    further down in the specification), child builders for each definition must be kept
    around until the end.

    Arguments:
        compdef_builder: Parent component definition builder.
        pos_tok: Position of the start of the component definition.
        name_tok: Name of the component definition if it exists
        varparam_counter: Object for handing out unique numbers to elementary var/param
            nodes.
    """

    def __init__(
        self,
        compdef_builder: "ComponentDefBuilder",
        pos_tok: "Token",
        name_tok: Optional["Token"],
        varparam_counter: Counter,
    ):
        self.diag_store: diagnostics.DiagnosticStore = compdef_builder.diag_store

        self.pos_tok = pos_tok
        self.name_tok = name_tok
        self.elements_by_label = collections.defaultdict(list)

        # Builders for specific parts of the component.
        self.varparam_builder = CompDefVarParamBuilder(self, varparam_counter)
        self.vargroup_builder = CompDefVarGroupBuilder(self)
        self.compinst_builder = CompDefCompInstBuilder(self)
        self.relinst_builder = CompDefRelInstBuilder(self)
        self.goal_builder = CompDefGoalBuilder(self)
        self.transform_builder = CompDefTransformBuilder(self)
        self.behavior_builder = CompDefBehaviorBuilder(self)
        self.design_builder = CompDefDesignBuilder(self)
        self.need_builder = CompDefNeedBuilder(self)
        self.comment_builder = CompDefCommentBuilder(self)

    def add_variables(self, new_vars: List[components.VarParam]):
        """Forward call to variable & parameter builder."""
        self.varparam_builder.add_variables(new_vars)

    def add_parameters(self, new_params: List[components.VarParam]):
        """Forward call to variable & parameter builder."""
        self.varparam_builder.add_parameters(new_params)

    def new_vargroup(self, name_tok: "Token"):
        """Forward call to variable group builder."""
        self.vargroup_builder.new_vargroup(name_tok)

    def vgroup_add_vars(self, varname_toks: List["Token"]):
        """Forward call to variable group builder."""
        self.vargroup_builder.vgroup_add_vars(varname_toks)

    def add_compinst(self, inst_name_tok: "Token", def_name_tok: "Token", has_arguments: bool):
        """Forward call to component instance builder."""
        self.compinst_builder.add_compinst(inst_name_tok, def_name_tok, has_arguments)

    def add_compinst_arguments(self, arguments: List["Token"]):
        """Forward call to component instance builder."""
        self.compinst_builder.add_compinst_arguments(arguments)

    def new_relinst(self, inst_name_tok: "Token", def_name_tok: "Token"):
        """Forward call to relation instance builder."""
        self.relinst_builder.new_relinst(inst_name_tok, def_name_tok)

    def relinst_argheader(self, argkind_tok: "Token"):
        """Forward call to relation instance builder."""
        self.relinst_builder.relinst_argheader(argkind_tok)

    def add_relinst_arguments(self, name_toks: List["Token"]):
        """Forward call to relation instance builder."""
        self.relinst_builder.add_relinst_arguments(name_toks)

    def new_goal_header(self, goal_kind: "Token"):
        """Forward call to goal builder."""
        self.goal_builder.new_goal_header(goal_kind)

    def add_goal(self, goal: components.Goal):
        """Forward call to goal builder."""
        self.goal_builder.add_goal(goal)

    def add_goal_subclause(self, sub_clause: components.SubClause):
        """Forward call to goal builder."""
        self.goal_builder.add_goal_subclause(sub_clause)

    def new_transform_header(self, transform_kind: "Token"):
        """Forward call to transform builder."""
        self.transform_builder.new_transform_header(transform_kind)

    def add_transform(self, transform: components.Transformation):
        """Forward call to transform builder."""
        self.transform_builder.add_transform(transform)

    def add_transform_subclause(self, sub_clause: components.SubClause):
        """Forward call to transform builder."""
        self.transform_builder.add_transform_subclause(sub_clause)

    def new_behavior_header(self, kind_tok: "Token"):
        """Forward call to behavior builder."""
        self.behavior_builder.new_behavior_header(kind_tok)

    def new_behavior_function(self, label_tok: "Token"):
        """Forward call to behavior builder."""
        self.behavior_builder.new_behavior_function(label_tok)

    def behavior_case(self, case_label_tok: "Token"):
        """Forward call to behavior builder."""
        self.behavior_builder.behavior_case(case_label_tok)

    def behavior_normal_when(self, when_tok: "Token"):
        """Forward call to behavior builder."""
        self.behavior_builder.behavior_normal_when(when_tok)

    def behavior_default_when(self, when_tok: "Token"):
        """Forward call to behavior builder."""
        self.behavior_builder.behavior_default_when(when_tok)

    def behavior_when_condition(
        self,
        name_tok: "Token",
        condition: Union["exprs.Disjunction", "exprs.RelationComparison"],
    ):
        """Forward call to behavior builder."""
        self.behavior_builder.behavior_when_condition(name_tok, condition)

    def behavior_normal_then(self, then_tok: "Token"):
        """Forward call to behavior builder."""
        self.behavior_builder.behavior_normal_then(then_tok)

    def behavior_then_result(self, name_tok: "Token", result: "exprs.Comparison"):
        """Forward call to behavior builder."""
        self.behavior_builder.behavior_then_result(name_tok, result)

    def new_design_header(self, kind: "Token"):
        """Forward call to design builder."""
        self.design_builder.new_design_header(kind)

    def design_line(self, design: components.Design):
        """Forward call to design builder."""
        self.design_builder.design_line(design)

    def add_design_subclause(self, sub: components.SubClause):
        """Forward call to design builder."""
        self.design_builder.add_design_subclause(sub)

    def add_need(self, label_tok: "Token", subject_tok: "Token", description: str):
        """Forward call to need builder."""
        self.need_builder.add_need(label_tok, subject_tok, description)

    def add_comment(self, name_tok: "Token"):
        """Forward call to comment builder."""
        self.comment_builder.add_comment(name_tok)

    def get_child_compdefnames(self) -> Set[str]:
        """Get the names of component definitions that are needed for child instances.

        Returns:
            Names of the components being instantiated in this component.
        """
        return self.compinst_builder.get_compdef_names()

    def finish(self, spec: "Specification", doc_distributor: "DocCommentDistributor"):
        """Parsing is finished, component child instances have been checked already.
        Check the collected component data, and add the component definition to the
        specification.

        Arguments:
            spec: Specification to use as source for types, verbs relation definitions,
                and other component definitions, and to fill with the type-checked
                component.
            doc_distributor: Object that accepts the found doc comments for distributing
                them to the elements of the specification.
        """
        comp_def = components.ComponentDefinition(self.pos_tok, self.name_tok)
        self.varparam_builder.finish_comp(comp_def, spec)
        self.vargroup_builder.finish_comp(comp_def, spec)
        self.compinst_builder.finish_comp(comp_def, spec)
        self.relinst_builder.finish_comp(comp_def, spec)
        self.goal_builder.finish_comp(comp_def, spec)
        self.transform_builder.finish_comp(comp_def, spec)
        self.behavior_builder.finish_comp(comp_def, spec)
        self.design_builder.finish_comp(comp_def, spec)
        self.need_builder.finish_comp(comp_def, spec)
        self.comment_builder.finish_comp(comp_def, doc_distributor)  # Must be final build step.

        # Check for unique labels within the scope of each comp_def.
        for labeled_elements in self.elements_by_label.values():
            if len(labeled_elements) > 1:
                locs = []
                for elem in labeled_elements:
                    if isinstance(elem, (components.VarParam, CollectedVarGroup)):
                        locs.append(elem.name_tok.get_location())
                    elif isinstance(elem, ParsedBehavior):
                        locs.append(elem.name.get_location())
                    elif isinstance(elem, (RelInst, ComponentInstance)):
                        locs.append(elem.inst_name_tok.get_location())
                    else:
                        locs.append(elem.label_tok.get_location())

                if self.name_tok:
                    scope = "component definition {}".format(self.name_tok.tok_text)
                else:
                    scope = "world"

                if hasattr(labeled_elements[-1], "label_tok"):
                    name = labeled_elements[-1].label_tok.tok_text
                elif hasattr(labeled_elements[-1], "name_tok"):
                    name = labeled_elements[-1].name_tok.tok_text
                elif hasattr(labeled_elements[-1], "name"):
                    name = labeled_elements[-1].name.tok_text
                elif hasattr(labeled_elements[-1], "inst_name_tok"):
                    name = labeled_elements[-1].inst_name_tok.tok_text
                else:
                    name = ""

                self.diag_store.add(
                    diagnostics.E227(
                        name,
                        scope,
                        location=locs[0],
                        dupes=locs,
                    )
                )

        # Add all elements of the component to the documentation distributor.
        for elem in components.get_doc_comment_comp_elements(comp_def):
            doc_distributor.add_element(elem)

        if comp_def.name_tok is None:
            spec.world = comp_def
        else:
            spec.comp_defs.append(comp_def)

    def notify_parameter_section(self, pos_tok: "Token"):
        """A parameter section was found, check if it is allowed."""
        if self.name_tok is None:  # We're processing 'world'
            # Parameter section not allowed in 'world'.
            self.diag_store.add(diagnostics.E201("parameter", "world", pos_tok.get_location()))

    def notify_transform_section(self, pos_tok: "Token"):
        """A transform section was found, check if it is allowed."""
        if self.name_tok is None:  # We're processing 'world'
            # Transformation section not allowed in 'world'.
            self.diag_store.add(diagnostics.E201("transformation", "world", pos_tok.get_location()))


class ComponentDefBuilder:
    """Builder to construct component definitions of the entire specification. The
    builder keeps a list of component child builders, one for each component definition.
    The latter do all the work for each component definition.
    """

    def __init__(self, ast_builder: "AstBuilder"):
        self.diag_store: diagnostics.DiagnosticStore = ast_builder.diag_store
        self.varparam_counter = Counter(100)

        ast_builder.register_new_section(self)
        self.child_builders: List[CompDefChildBuilders] = []
        self.current_component: Optional[CompDefChildBuilders] = None

    def notify_new_section(self, new_top_section):
        """Notification for self and possibly the child builders if a new component
        definition is under construction.
        """
        if new_top_section:
            self.current_component = None

    def new_componentdef(self, pos_tok: "Token", name_tok: Optional["Token"]):
        """New component definition started.

        Arguments:
            pos_tok: Token defining the start position of the component definition.
            name_tok: Token with the name of the definition if it exists.
                Non-existing name means the component represents 'world'.
        """
        self.current_component = CompDefChildBuilders(
            self, pos_tok, name_tok, self.varparam_counter
        )
        self.child_builders.append(self.current_component)

    def finish(self, spec: "Specification", doc_distributor: "DocCommentDistributor"):
        """Parsing has finished, complete type checking.

        Arguments:
            spec: Specification already containing types, verb-prepositions, and
                relation definitions. Must be filled with component definitions and
                world component.
            doc_distributor: Object that accepts the found doc comments for distributing
                them to the elements of the specification.
        """
        # Verify that component definitions are sufficiently unique.
        compdefs: Dict[str, List[CompDefChildBuilders]] = collections.defaultdict(list)
        worlds = []
        for ch_builder in self.child_builders:
            if ch_builder.name_tok is None:
                worlds.append(ch_builder)
            else:
                name = ch_builder.name_tok.tok_text
                compdefs[name].append(ch_builder)

        if len(worlds) == 0:
            self.diag_store.add(diagnostics.E202("world definition"))
        elif len(worlds) > 1:
            locs = [world.pos_tok.get_location() for world in worlds]
            self.diag_store.add(
                diagnostics.E200("world", "component instance", location=locs[0], dupes=locs)
            )

        for cdefs in compdefs.values():
            if len(cdefs) > 1:
                # Duplicate component definitions.
                locs = [cdef.pos_tok.get_location() for cdef in cdefs]
                assert cdefs[0].name_tok is not None
                name = cdefs[0].name_tok.tok_text
                self.diag_store.add(
                    diagnostics.E200(name, "component definition", location=locs[0], dupes=locs)
                )

        # Components to use.
        comp_builders = [cdef[0] for cdef in compdefs.values()]
        if len(worlds) > 0:
            comp_builders.append(worlds[0])

        # Order component definitions such that they are defined before they are used as
        # child component instances.
        orderer = Orderer()
        for comp_builder in comp_builders:
            child_defnames = comp_builder.get_child_compdefnames()
            if comp_builder.name_tok:
                compdef_name = comp_builder.name_tok.tok_text
            else:
                compdef_name = "world"  # Name cannot exist otherwise.

            orderer.add_dependency(compdef_name, child_defnames, comp_builder)

        resolveds, cycle = orderer.resolve()
        for entry in resolveds:
            if entry.data is None:
                # Entry was created by the orderer, we should run into it again and fail
                # to find it. Note that a component definition is always added to the
                # spec even if it contains errors. Therefore, presence of a component
                # definition in spec means it was present in the ESL text.
                continue

            entry.data.finish(spec, doc_distributor)

        if cycle:
            assert entry.data is not None
            locs = [entry.data.pos_tok.get_location() for entry in cycle]
            assert cycle[0].data is not None
            comp_name = cycle[0].data.name_tok.tok_text  # Cannot be world!
            self.diag_store.add(
                diagnostics.E204(comp_name, "component definition", location=locs[0], cycle=locs)
            )



"""Component definitions with their contents."""
from typing import TYPE_CHECKING, Generator, List, Optional, Union

from raesl.compile.ast import comment_storage
from raesl.utils import split_first_dot

if TYPE_CHECKING:
    from raesl.compile.ast.exprs import (
        Expression,
        Disjunction,
        RelationComparison,
        Comparison,
    )
    from raesl.compile.ast.nodes import VarNode, Node
    from raesl.compile.ast.relations import RelationDefinition
    from raesl.compile.scanner import Token
    from raesl.compile.ast.types import BaseType


# Kinds of goals and transformations.
CONSTRAINT = "constraint"
REQUIREMENT = "requirement"


class ComponentDefinition(comment_storage.DefaultDocStore):
    """ESL component definition.

    Arguments:
        pos_tok: Position of the definition. Either the name token or the 'world' token.
        name_tok: Token with the name of the component definition, None means 'world'.

    Attributes:
        variables: Variables of the component definition.
        parameters: Parameters of the component definition.
        var_groups: Groups of variables with a name.
        component_instances: Component instances of the component definition.
        needs: Needs of the component definition.
        goals: Goals of the component definition.
        transforms: Transformations of the component definition.
        designs: Designs of the component definition.
        relations: Relation instances of the component definition.
        behaviors: Behavior functions of the component definition.
    """

    def __init__(self, pos_tok: "Token", name_tok: Optional["Token"]):
        super(ComponentDefinition, self).__init__(name_tok)
        self.pos_tok = pos_tok
        self.name_tok = name_tok

        self.variables: List[VarParam] = []
        self.parameters: List[VarParam] = []
        self.var_groups: List[VariableGroup] = []
        self.component_instances: List[ComponentInstance] = []
        self.needs: List[Need] = []
        self.goals: List[Goal] = []
        self.transforms: List[Transformation] = []
        self.designs: List[Design] = []
        self.relations: List[RelationInstance] = []
        self.behaviors: List["BehaviorFunction"] = []


class VarParam(comment_storage.DocStore):
    """ESL component definition variable or parameter.

    Arguments:
        is_variable: Whether the object represents a variable.
        name_tok: Token with the name of the variable being defined.
        type_tok: Token with the name of the type of the variable being defined.
        is_property: Whether the parameter is a property.

    Attributes:
        type: Type of the variable, if it exists. Set during type checking.
    """

    def __init__(
        self,
        is_variable: bool,
        name_tok: "Token",
        type_tok: "Token",
        is_property: bool = False,
    ):
        super(VarParam, self).__init__(name_tok)
        self.is_variable = is_variable
        self.name_tok = name_tok
        self.type_tok = type_tok
        self.is_property = is_property
        self.type: Optional["BaseType"] = None

        self.node: Optional["VarNode"] = None

        # Variables are never explicit property, since they naturally belong
        # to the component defining them.
        assert not self.is_property or not self.is_variable

    def resolve_node(self, name: str) -> Optional["VarNode"]:
        """Find the varparam (sub) node that matches the dotted 'name'.

        Arguments:
            name: Possibly dotted name that should point at an existing sub-node.
                The empty string denotes 'self'.

        Returns:
            The node that matches the name, or None if no such node exists. In the
                latter case, use 'self.get_error_position(name)' to get
                an indication where the match fails in the name.
        """
        local_name, remaining_name, _dot_length = split_first_dot(name)

        if self.name_tok.tok_text != local_name:
            return None

        assert self.node is not None, "Trying to use non-existing node of '{}'".format(
            self.name_tok.tok_text
        )
        return self.node.resolve_node(remaining_name)

    def resolve_element(self, name: str) -> Optional[comment_storage.DocAddElement]:
        node = self.resolve_node(name)
        if node is None:
            return None
        else:
            assert isinstance(node, comment_storage.DocAddElement)
            return node

    def get_error_position(self, name: str):
        """Return the index in the given string where an error occurs in resolving the
        node.

        Arguments:
            name: Name of the element to find.

        Returns:
            Approximated index in the string where matching the element fails.
                Returned value has no meaning if resolving a node succeeds.
        """
        local_name, remaining_name, dot_length = split_first_dot(name)

        if self.name_tok.tok_text != local_name:
            return 0  # Local name == first name is wrong.
        else:
            # Ask child about the position of the error.
            offset = self.node.get_error_position(remaining_name)
            return offset + len(local_name) + dot_length

    def __repr__(self):
        if self.is_variable:
            kind = "Variable"
        else:
            kind = "Parameter"
        text = "{}[{}, {}, {}]"
        return text.format(kind, self.name_tok.tok_text, self.type_tok.tok_text, self.is_property)


class VariableGroup:
    """One variable group in ESL (a named group of variables).

    It has no documentation comment, as its only purpose is to enable interfacing
    to child components.

    As a variable group doesn't need to contain uniquely named variables, their names
    cannot be used to build a Compound type. Therefore, it just stays a group, and it
    gets dealt with in component instantiation.

    Arguments:
        name_tok: Token with the name of the group being defined.
        variablepart_names: Tokens with possibly dotted name of variable parts in the
            group.

    Attributes:
        node: Node representing the group, if available.
    """

    def __init__(self, name_tok: "Token", variablepart_names: List["Token"]):
        self.name_tok = name_tok
        self.variablepart_names = variablepart_names

        # Set during type checking.
        self.node: Optional["Node"] = None


class InstanceArgument:
    """Actual argument of a component or relation.

    Arguments:
        name_tok: Name of the actual argument.
        argnode: Node of the argument, filled during type checking.
    """

    def __init__(self, name_tok: "Token", argnode: Optional["Node"] = None):
        self.name_tok = name_tok
        self.argnode = argnode


class ComponentInstance(comment_storage.DefaultDocStore):
    """ESL component instance in a component definition.

    Arguments:
        inst_name_tok: Token with the name of the component instance.
        def_name_tok: Token withe the name of the component definition to apply.

    Attributes:
        arguments: Arguments of the instance.
        compdef: Component definition matching the name in 'def_name_tok', if it exists.
            Set during type checking.
    """

    def __init__(self, inst_name_tok: "Token", def_name_tok: "Token"):
        super(ComponentInstance, self).__init__(inst_name_tok)
        self.inst_name_tok = inst_name_tok
        self.def_name_tok = def_name_tok
        self.arguments: List[InstanceArgument] = []

        self.compdef: Optional[ComponentDefinition] = None


class RelationInstance(comment_storage.DefaultDocStore):
    """ESL relation instance in a component definition.

    Arguments:
        inst_name_tok: Token with the name of the relation instance.
        def_name_tok: Token withe the name of the relation definition to apply.
        arguments: Arguments of the instance. One element for each parameter, where one
            element may have several arguments due to the 'one or more' feature.
        reldef: Relation definition of this instance.
    """

    def __init__(
        self,
        inst_name_tok: "Token",
        def_name_tok: "Token",
        arguments: List[List[InstanceArgument]],
        reldef: Optional["RelationDefinition"],
    ):
        super(RelationInstance, self).__init__(inst_name_tok)
        self.inst_name_tok = inst_name_tok
        self.def_name_tok = def_name_tok
        self.arguments = arguments
        self.reldef = reldef


class Flow:
    """Flow in a goal or Transformation.

    Arguments:
        name_tok: Dotted name of the flow.

    Attributes:
        flow_node: If not None, node represented by the flow.
    """

    def __init__(self, name_tok):
        self.name_tok = name_tok
        self.flow_node: Optional[VarNode] = None


class Goal(comment_storage.DefaultDocStore):
    """Goal in an ESL component definition.

    Arguments:
        label_tok: Label name of the goal.
        active: Token with the name of the active component.
        doesaux: 'does' or auxiliary word token.
        verb: Verb of the goal.
        flows: Flows of the goal.
        prepos: Token with the preposition word.
        passive: Token with the name of the passive component.

    Attributes:
        goal_kind: Kind of goal, filled after construction.
            Either 'requirement' or 'constraint' string.
        sub_clauses: Sub-clauses of the goal.
        active_comp: If not None, resolved active component instance of the goal.
        passive_comp: If not None, resolved passive component instance of the goal.
    """

    def __init__(
        self,
        label_tok: "Token",
        active: "Token",
        doesaux: "Token",
        verb: "Token",
        flows: List[Flow],
        prepos: "Token",
        passive: "Token",
    ):
        super(Goal, self).__init__(label_tok)
        self.goal_kind: Optional[str] = None
        self.label_tok = label_tok
        self.active = active
        self.doesaux = doesaux
        self.verb = verb
        self.flows = flows
        self.prepos = prepos
        self.passive = passive
        self.sub_clauses: List[SubClause] = []

        self.active_comp: Optional[ComponentInstance] = None
        self.passive_comp: Optional[ComponentInstance] = None


class Transformation(comment_storage.DefaultDocStore):
    """Transformation in a component.

    Arguments:
        label_tok: Label name of the transformation.
        doesaux_tok: 'does' or aux word token.
        verb_tok: Verb of the transformation.
        in_flows: Inputs required for the transformation.
        prepos_tok: Preposition of the transformation.
        out_flows: Outputs resulting from the transformation.

    Attributes:
        transform_kind: Kind of transformation, filled after construction.
            Either 'requirement' or 'constraint' string.
        sub_clauses: Sub-clauses of the transformation.
    """

    def __init__(
        self,
        label_tok: "Token",
        doesaux_tok: "Token",
        verb_tok: "Token",
        in_flows: List[Flow],
        prepos_tok: "Token",
        out_flows: List[Flow],
    ):
        super(Transformation, self).__init__(label_tok)
        self.transform_kind: Optional[str] = None
        self.label_tok = label_tok
        self.doesaux_tok = doesaux_tok
        self.verb_tok = verb_tok
        self.in_flows = in_flows
        self.prepos_tok = prepos_tok
        self.out_flows = out_flows
        self.sub_clauses: List[SubClause] = []


class Design(comment_storage.DefaultDocStore):
    """Design rule in a component.

    Arguments:
        label_tok: Name of the design rule.
        expr: Condition expressed in the design.

    Attributes:
        design_kind: Kind of the design, filled in after construction.
            Contains either 'requirement' or 'constraint'.
        sub_clauses: Sub-clauses of the design.
    """

    def __init__(self, label_tok: "Token", expr: "Expression"):
        super(Design, self).__init__(label_tok)
        self.design_kind: Optional[str] = None
        self.label_tok = label_tok
        self.expr = expr
        self.sub_clauses: List[SubClause] = []


class SubClause:
    """Subclause in a goal, transformation, or behavior.

    Arguments:
        label_tok: Name of the subclause.
        expr: Expression describing the subclause.
    """

    def __init__(self, label_tok: "Token", expr: "Expression"):
        self.label_tok = label_tok
        self.expr = expr


class BehaviorFunction(comment_storage.DefaultDocStore):
    """One function specifying some behavior.

    Arguments:
        behavior_kind: Kind of behavior. Either 'requirement' or 'constraint'.
        name_tok: Name of the behavior.

    Attributes:
        cases: Behavior cases.
        default_results: Results that hold when none of the cases applies. None means
            there is no default result.
    """

    def __init__(self, behavior_kind: str, name_tok: "Token"):
        super(BehaviorFunction, self).__init__(name_tok)
        self.behavior_kind = behavior_kind
        self.name_tok = name_tok
        self.cases: List[BehaviorCase] = []
        self.default_results: Optional[List[BehaviorResult]] = None


class BehaviorCondition:
    """A condition of a case."""

    def __init__(self, name_tok: "Token", comparison: Union["Disjunction", "RelationComparison"]):
        self.name_tok = name_tok
        self.comparison = comparison


class BehaviorResult:
    """A result of a case."""

    def __init__(self, name_tok: "Token", result: "Comparison"):
        self.name_tok = name_tok
        self.result = result


class BehaviorCase:
    """A set of desired behavioral results given a set of conditions.

    Arguments:
        name_tok: Name of the behavior case.
        conditions: Conditions that should hold for the case to apply.
        results: Results that should hold when the case applies.
    """

    def __init__(
        self,
        name_tok: "Token",
        conditions: List[BehaviorCondition],
        results: List[BehaviorResult],
    ):
        self.name_tok = name_tok
        self.conditions = conditions
        self.results = results


NeedSubjectTypes = Union[
    RelationInstance,
    VarParam,
    ComponentInstance,
    Goal,
    Transformation,
    Design,
    BehaviorFunction,
]


class Need(comment_storage.DefaultDocStore):
    """Informal need in ESL.

    Arguments:
        label_tok: Token with the name of the label.
        subject_tok: Token with the name of the subject of the need.
        description: Description of the need.

    Attributes:
        subject: If not None, subject of the need.
    """

    def __init__(self, label_tok: "Token", subject_tok: "Token", description: str):
        super(Need, self).__init__(label_tok)
        self.label_tok = label_tok
        self.subject_tok = subject_tok
        self.description = description

        self.subject: Optional[NeedSubjectTypes] = None


def get_doc_comment_comp_elements(
    comp: ComponentDefinition,
) -> Generator[comment_storage.DocStore, None, None]:
    """Retrieve the component elements interested in getting documentation comments from
    the input. This includes the component itself, so you can add documentation to it
    in its 'comments' section.

    Arguments:
        comp: Component definition to search.

    Returns:
        Generator yielding interested elements.
    """
    all_elems: List[
        Union[
            List[ComponentDefinition],
            List[VarParam],
            List[ComponentInstance],
            List[Need],
            List[Goal],
            List[Transformation],
            List[Design],
            List[RelationInstance],
            List[BehaviorFunction],
        ]
    ] = [
        [comp],
        comp.variables,
        comp.parameters,
        comp.component_instances,
        comp.needs,
        comp.goals,
        comp.transforms,
        comp.designs,
        comp.relations,
        comp.behaviors,
    ]
    for elems in all_elems:
        for elem in elems:
            if elem.doc_tok:
                yield elem



"""Classes to collect and store information from the parsing process, perform checking
on the information for being correct, and construct an AST as result along with a log of
found diagnostics.

The AstBuilder operates at top section level (types, verbs, relation definitions, and
component definitions). It leaves all the details of each section to dedicated child
builders (thus creating a highly modular checker), and acts as call dispatcher and
global controller in the type-checking and ast building process once parsing has
finished.

Notable parts in the class are
- Child builders for each top section part.
- Diagnostics store shared with all the child builders.
- Storage of doc-comments in the input for attaching them to the correct parts of the
  produced specification after parsing.
- Call dispatcher for the child builders that a new top or sub-section has been found,
  allowing them to clean up processing if needed.
- Entry points for the parser to push found information to the child builders.
- The 'finish_parse' entry point to perform all type checking, and produce the AST and
  found diagnostics.
"""
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from raesl.compile import diagnostics
from raesl.compile.ast import exprs, specification
from raesl.compile.ast.comment_storage import DocCommentDistributor
from raesl.compile.typechecking.compdef_builder import ComponentDefBuilder
from raesl.compile.typechecking.reldef_builder import RelationDefBuilder
from raesl.compile.typechecking.type_builder import TypeBuilder
from raesl.compile.typechecking.verb_builder import VerbDefBuilder

if TYPE_CHECKING:
    from raesl.compile.scanner import Token


class AstBuilder:
    """Builder to collect information from the parse process, perform type
    checking, and produce an AST and reported diagnostics.

    Arguments:
        diag_store: Storage for diagnostics while building the AST.

    Attributes:
        doc_distributor: Object that distributes doc comments to interested elements of
           the specification.
        section_notify_list: Builders to notify of a new section.
        type_builder: Builder for constructing types.
        verb_builder: Builder for constructing verb/prepositions.
        reldef_builder: Builder for constructing relations.
        compdef_builder: Builder for constructing components.
    """

    def __init__(self, diag_store: diagnostics.DiagnosticStore):
        self.diag_store = diag_store
        self.doc_distributor = DocCommentDistributor(self.diag_store)
        self.section_notify_list: List[
            Union[TypeBuilder, RelationDefBuilder, ComponentDefBuilder]
        ] = []

        self.type_builder = TypeBuilder(self)
        self.verb_builder = VerbDefBuilder(self)
        self.reldef_builder = RelationDefBuilder(self)
        self.compdef_builder = ComponentDefBuilder(self)

    def add_typedef(
        self,
        type_name: "Token",
        parent_name: Optional["Token"],
        enum_spec: Optional[List[exprs.Value]],
        unit_spec: Optional[List["Token"]],
        ival_spec: Optional[List[Tuple[Optional[exprs.Value], Optional[exprs.Value]]]],
        cons_spec: Optional[exprs.Value],
    ):
        """Forward call to type builder."""
        self.type_builder.add_typedef(
            type_name, parent_name, enum_spec, unit_spec, ival_spec, cons_spec
        )

    def new_bundle_type(self, bundle_name: "Token"):
        """Forward call to type builder."""
        self.type_builder.new_bundle_type(bundle_name)

    def add_bundle_field(
        self,
        field_name: "Token",
        type_name: Optional["Token"],
    ):
        """Forward call to type builder."""
        self.type_builder.add_bundle_field(
            field_name,
            type_name,
        )

    def add_verbdef(self, verb_tok: "Token", prepos_tok: "Token"):
        """Forward call to verb definition builder."""
        self.verb_builder.add_verbdef(verb_tok, prepos_tok)

    def add_reldef(self, name: "Token"):
        """Forward call to relation definition builder."""
        self.reldef_builder.add_reldef(name)

    def reldef_param_header(self, header_tok: "Token", direction: str):
        """Forward call to relation definition builder."""
        self.reldef_builder.reldef_param_header(header_tok, direction)

    def reldef_add_param(self, name: "Token", type_name: "Token", multi: bool):
        """Forward call to relation definition builder."""
        self.reldef_builder.reldef_add_param(name, type_name, multi)

    def register_new_section(self, other_builder):
        """Entry point for a child builder to declare interest in receiving
        notifications about new sections in the file.
        """
        self.section_notify_list.append(other_builder)

    def notify_new_section(self, tok: Optional["Token"], new_top_section: bool):
        """Parser has started a new section, finish all 'in-progress' definitions.

        Arguments:
            tok: Token indicating the position of the new section. None is used for EOF.
            new_top_section: If set, a new type, verbs, or component definition has
                been found, else a new section within a component has been detected.
        """
        if tok:
            # New section started, documentation after this point doesn't belong
            # to a previous element.
            self.doc_distributor.add_dummy_element(tok)

        for builder in self.section_notify_list:
            builder.notify_new_section(new_top_section)

    def finish_parse(self, doc_comments: List["Token"]) -> Optional[specification.Specification]:
        """Finish processing the collected information, that is, perform type checking.

        Arguments:
            doc_comments: Raw documentation comments rescued from the scanner.
        """
        # Tell all builders current section is done.
        self.notify_new_section(None, True)

        # Convert collected information to AST.
        spec = specification.Specification()
        self.verb_builder.finish(spec)
        self.type_builder.finish(spec)
        self.reldef_builder.finish(spec)  # Requires types.
        self.compdef_builder.finish(spec, self.doc_distributor)  # Requires types, verbs, reldefs.

        # Add specification elements to the doc distributor.
        for elem in specification.get_doc_comment_spec_elements(spec):
            self.doc_distributor.add_element(elem)

        # Hand out all doc comments.
        self.doc_distributor.resolve(doc_comments)
        return spec



"""Functions for instantiating the component tree."""
from typing import Optional

from raesl.compile import diagnostics
from raesl.compile.ast.specification import Specification
from raesl.compile.instantiating.edge_building import EdgeFactory
from raesl.compile.instantiating.node_building import NodeFactory
from ragraph.graph import Graph


class GraphFactory:
    """Graph factory class.

    Converts a specification into a graph containing a node hierarchy and edges for
    derived dependencies between nodes.

    Arguments:
        diag_store: Storage for found diagnostics during the process.

    Attributes:
        node_factory: Factory that parses spec into Node objects.
        edge_factory: Factory that derives edges from found Node objects.
    """

    def __init__(
        self,
        diag_store: diagnostics.DiagnosticStore,
        spec: Optional[Specification] = None,
    ):
        self.diag_store = diag_store
        self.spec = spec
        self.node_factory = NodeFactory(self.diag_store, spec=spec)
        self.edge_factory = EdgeFactory(self.diag_store, node_store=self.node_factory.node_store)

    def make_graph(self, spec: Optional[Specification] = None) -> Optional[Graph]:
        """Instantiate the tree defined in the specification, and build a graph for it.

        Arguments:
            spec: Specification object holding parsed ESL data.

        Returns:
            None if no root is available, else the constructed graph.

        Note:
            Problems may be reported during instantiation and added to self.diag_store.
        """
        self.spec = self.spec if spec is None else spec
        if self.spec is None or self.spec.world is None:
            return None

        node_dict = self.node_factory.make_nodes(self.spec)
        edges = self.edge_factory.make_edges()

        return Graph(nodes=node_dict.values(), edges=edges)



"""Line matcher state machines of relation definitions."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import typing
from raesl.compile.typechecking import reldef_builder

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_RELATION_DEF_SPEC = """
relation_def:
    start initial;
    start -> s1 [DEFINE_KW] tag=define;
    s1 -> s2 [RELATION_KW];

    end accept=define_relation;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def _process_reldef_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["define"][0], True)


_RELATION_NAME_LINE_SPEC = """
relation_name:
    start initial;
    start -> s1 [NAME] tag=rel_name;

    end accept=relation_name;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_reldef_new_relation(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    builder.add_reldef(tags["rel_name"][0])


_RELATION_PARAMETER_HEADER_SPEC = """
relation_param_header_spec:
    start initial;
    start -> s1 [REQUIRING_KW] tag=rel_requiring;
    start -> s1 [RETURNING_KW] tag=rel_returning;
    start -> s1 [RELATING_KW] tag=rel_relating;
    s1 -> s2 [PARAMETER_KW];

    end accept=relation_param_header;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def _process_reldef_param_header(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    if "rel_requiring" in tags:
        token = tags["rel_requiring"][0]
        direction = reldef_builder.INPUT
    elif "rel_returning" in tags:
        token = tags["rel_returning"][0]
        direction = reldef_builder.OUTPUT
    else:
        assert "rel_relating" in tags
        token = tags["rel_relating"][0]
        direction = reldef_builder.INPOUT

    builder.reldef_param_header(token, direction)


_RELATION_PARAMETER_LINE_SPEC = """
relation_param:
    start initial;
    start -> s1 [STAR_TK];
    s1 -> s5 [NAME] tag=param_name;

    s1 -> s2 [ONE_KW] tag=multi_param;
    s2 -> s3 [OR_KW];
    s3 -> s4 [MORE_KW];

    s4 -> s5 [NAME] tag=param_name;
    s5 -> s6 [IS_KW];
    s6 -> s7 [A_KW];
    s7 -> s8 [NAME] tag=param_type;

    end accept=relation_param;
    s8 -> end [NL_TK];
    s8 -> end [EOF_TK];
"""


def _process_reldef_param_entry(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    param_name = tags["param_name"][0]
    param_type = tags["param_type"][0]
    builder.reldef_add_param(param_name, param_type, "multi_param" in tags)


MACHINES: typing.MachineTripletList = [
    ("DEFINE_RELATION_MACHINE", _RELATION_DEF_SPEC, _process_reldef_header),
    (
        "RELATION_NAME_LINE_MACHINE",
        _RELATION_NAME_LINE_SPEC,
        _process_reldef_new_relation,
    ),
    (
        "RELATION_PARAMETER_HEADER_MACHINE",
        _RELATION_PARAMETER_HEADER_SPEC,
        _process_reldef_param_header,
    ),
    (
        "RELATION_PARAMETER_LINE_MACHINE",
        _RELATION_PARAMETER_LINE_SPEC,
        _process_reldef_param_entry,
    ),
]



"""Support functions."""
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

from raesl.compile import diagnostics
from raesl.compile.ast.nodes import VarNode
from raesl.utils import get_first_namepart

if TYPE_CHECKING:
    from raesl.compile.ast.components import (
        ComponentDefinition,
        VarParam,
        VariableGroup,
    )
    from raesl.compile.ast.components import (
        ComponentInstance,
        RelationInstance,
        Goal,
        Transformation,
        Design,
        BehaviorFunction,
    )
    from raesl.compile.ast.nodes import Node
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token


def split_arguments(
    params_length: int, multiple_index: Optional[int], arguments: List["Token"]
) -> List[List["Token"]]:
    """Given a list arguments, split them into 'params_length' pieces, where each piece
    has length 1, except piece 'multiple_index' if not None, which takes all the slack.

    Arguments:
        params_length: Number of pieces in the result.
        multiple_index: Index in 'arguments' where the multi-piece starts, only if
            multiple_index is not None.
        arguments: Actual arguments to split in pieces.
    """
    if multiple_index is None:
        assert len(arguments) == params_length
        return [[arg] for arg in arguments]

    assert len(arguments) >= params_length
    assert multiple_index >= 0
    assert multiple_index < params_length
    num_singular = params_length - 1
    length_multi = len(arguments) - num_singular
    after_multiple = multiple_index + length_multi

    return (
        [[arg] for arg in arguments[:multiple_index]]
        + [arguments[multiple_index:after_multiple]]
        + [[arg] for arg in arguments[after_multiple:]]
    )


def construct_var_param_map(comp_def: "ComponentDefinition") -> Dict[str, "VarParam"]:
    """Construct a dict of variable / parameter names to their definitions.

    Arguments:
        comp_def: Definition to search for available variables and parameters.

    Returns:
        Dictionary of names to their definitions.
    """
    vps: Dict[str, "VarParam"] = {}  # Map of name to definition.
    for var in comp_def.variables:
        vps[var.name_tok.tok_text] = var
    for param in comp_def.parameters:
        vps[param.name_tok.tok_text] = param
    return vps


def construct_vargroup_map(
    comp_def: "ComponentDefinition",
) -> Dict[str, "VariableGroup"]:
    """Construct a dict of variable groups names to their definitions.

    Arguments:
        comp_def: Definition to search for available variables and parameters.

    Returns:
        Dictionary of group names to their definitions.
    """
    return dict((var.name_tok.tok_text, var) for var in comp_def.var_groups)


def construct_comp_instances_map(
    comp_def: "ComponentDefinition",
) -> Dict[str, "ComponentInstance"]:
    """
    Construct a dict of child component instance names of the given
    component definition.

    :param comp_def: Definition to search for available component instances.
    :return: Dictionary of component instance names to their instances.
    """
    compinsts = {}
    for cinst in comp_def.component_instances:
        compinsts[cinst.inst_name_tok.tok_text] = cinst
    return compinsts


def construct_verb_prepos_combis(spec: "Specification") -> Set[Tuple[str, str]]:
    """
    Construct a set with all defined verb/prepos combinations.
    """
    return set((vpp.verb.tok_text, vpp.prepos.tok_text) for vpp in spec.verb_prepos)


def construct_relinst_goal_transform_design_behavior_map(
    comp_def: "ComponentDefinition",
) -> Dict[str, Union["RelationInstance", "Goal", "Transformation", "Design", "BehaviorFunction"],]:
    """Construct a dict to quickly find goals, transformations, designs, and behaviors
    by their label name.

    Arguments:
        comp_def: Definition to search.

    Returns:
        Dictionary of labels to their goals, transformations, designs, and behaviors.
    """
    label_map: Dict[
        str,
        Union["RelationInstance", "Goal", "Transformation", "Design", "BehaviorFunction"],
    ]
    label_map = {}
    for relinst in comp_def.relations:
        label_map[relinst.inst_name_tok.tok_text] = relinst
    for goal in comp_def.goals:
        label_map[goal.label_tok.tok_text] = goal
    for trans in comp_def.transforms:
        label_map[trans.label_tok.tok_text] = trans
    for design in comp_def.designs:
        label_map[design.label_tok.tok_text] = design
    for behavior in comp_def.behaviors:
        label_map[behavior.name_tok.tok_text] = behavior
    return label_map


def resolve_var_param_node(
    name_tok: "Token",
    avail_vps: Dict[str, "VarParam"],
    reported_names: Set[str],
    diag_store: diagnostics.DiagnosticStore,
) -> Optional["VarNode"]:
    """Resolve the (possibly sub)node of a variable or parameter indicated by 'name'.
    If it fails, report an error if necessary.

    Arguments:
        name_tok: Name of the node to obtain, may contain a dotted name.
        avail_vps: Variables and parameters available in the context.
        reported_names: Non-existing names and prefixes that are reported already.
        diag_store: Storage for found diagnostics.

    Returns:
        The node represented by the name, or None if it could not be found.
            In the latter case, an problem exists indicating failure to find the node.
    """
    node = resolve_var_param_group_node(name_tok, avail_vps, None, reported_names, diag_store)
    if node is None:
        return None
    assert isinstance(node, VarNode)
    return node


def resolve_var_param_group_node(
    name_tok: "Token",
    avail_vps: Optional[Dict[str, "VarParam"]],
    avail_vgroups: Optional[Dict[str, "VariableGroup"]],
    reported_names: Set[str],
    diag_store: diagnostics.DiagnosticStore,
) -> Optional["Node"]:
    """Resolve the provided (possibly dotted) name to a node from a variable, parameter
    or variable group.

    Arguments:
        name_tok: Name of the node to obtain, may contain a dotted name.
        avail_vps: Available variables and parameters, may be None.
        avail_vgroups: Available variable groups, may be None.
        reported_names: Non-existing variables, parameters, and groups that are
            reported already.
        diag_store: Destination for found diagnostics.

    Returns:
        The node represented by the name. It can be a Node if the name points at a
            variable groups. It is always a VarNode if the nam points at a variable or
            parameter.
    """
    first_part = get_first_namepart(name_tok.tok_text)
    if avail_vgroups is not None:
        vgrp = avail_vgroups.get(first_part)
        if vgrp is not None:
            i = name_tok.tok_text.find(".")
            if i >= 0:
                # It is a dotted name, which is not allowed in a group.
                # Report an error if not done already.
                if first_part not in reported_names:
                    reported_names.add(first_part)
                    diag_store.add(
                        diagnostics.E224(
                            "variable group",
                            f"selections like '{name_tok.tok_text[i:]}'",
                            location=name_tok.get_location(len(first_part)),
                        )
                    )
                    return None

            return vgrp.node

    # No variable groups, or no match, try variable or parameters.
    if avail_vps is not None:
        varparam = avail_vps.get(first_part)
        if varparam is not None:
            node = varparam.resolve_node(name_tok.tok_text)
            if node is not None:
                return node

            # varparam.resolve_node failed. Do report multiple times for the same
            # var/param for different dotted suffixes.
            if name_tok.tok_text not in reported_names:
                reported_names.add(name_tok.tok_text)

                kind = {True: "variable", False: "parameter"}[varparam.is_variable]
                offset = varparam.get_error_position(name_tok.tok_text)
                # Cannot resolve part of a dotted name.
                diag_store.add(
                    diagnostics.E225(
                        name_tok.tok_text[offset:],
                        first_part,
                        kind,
                        location=name_tok.get_location(offset),
                    )
                )

            return None

    # Name does not exist.
    if first_part not in reported_names:
        reported_names.add(first_part)

        if avail_vps is not None:
            if avail_vgroups is not None:
                kind = "variable, parameter, or variable group instance"
            else:
                kind = "variable or parameter instance"
        else:
            assert avail_vgroups is not None, "Must have at least one set of names."
            kind = "variable group instance"
        diag_store.add(diagnostics.E203(kind, name=first_part, location=name_tok.get_location()))

    return None



"""RaESL plotting and visualization module."""

from raesl.plot.diagrams import function_chain_diagram  # noqa
from raesl.plot.diagrams import function_traceability_diagram  # noqa
from raesl.plot.diagrams import functional_context_diagram  # noqa
from raesl.plot.diagrams import functional_dependency_diagram  # noqa
from raesl.plot.diagrams import hierarchy_diagram  # noqa
from raesl.plot.generic import Style  # noqa
from raesl.plot.matrix import mdm  # noqa



"""Line matchers for component instances within world or a component definition."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import argument_list, typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_COMPONENT_HEADER_SPEC = """
component_header:
    start initial;
    start -> s1 [COMPONENT_KW] tag=component;

    end accept=component_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_component_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["component"][0], False)


_COMPONENT_INSTANCE_NO_ARGS_SPEC = """
component_instance:
    start initial;
    start -> s1 [NAME] tag=component_instance;
    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [NAME] tag=component_def;

    end accept=component_instance_no_arguments;
    s4 -> end [NL_TK];
    s4 -> end [EOF_TK];
"""


def _process_instance_no_args(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    instance_tok = tags["component_instance"][0]
    definition_tok = tags["component_def"][0]
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_compinst(instance_tok, definition_tok, False)


_COMPONENT_INSTANCE_WITH_ARGS_SPEC = """
component_instance:
    start initial;
    start -> s1 [NAME] tag=component_instance;
    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [NAME] tag=component_def;

    s4 -> s5 [WITH_KW];
    s5 -> s6 [ARGUMENT_KW];

    end accept=component_instance_with_arguments;
    s6 -> end [NL_TK];
"""


def _process_instance_with_args(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    instance_tok = tags["component_instance"][0]
    definition_tok = tags["component_def"][0]
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_compinst(instance_tok, definition_tok, True)


# Argument line processing.
def _process_instance_arguments(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    arguments = argument_list.process_argument_list_line(tags)
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_compinst_arguments(arguments)


MACHINES: typing.MachineTripletList = [
    ("COMPONENT_HEADER_MACHINE", _COMPONENT_HEADER_SPEC, _process_component_header),
    (
        "COMPONENT_INSTANCE_NO_ARGS_MACHINE",
        _COMPONENT_INSTANCE_NO_ARGS_SPEC,
        _process_instance_no_args,
    ),
    (
        "COMPONENT_INSTANCE_WITH_ARGS_MACHINE",
        _COMPONENT_INSTANCE_WITH_ARGS_SPEC,
        _process_instance_with_args,
    ),
    (
        "COMPONENT_ARGUMENT_MACHINE",
        argument_list.ARGUMENT_LINE_SPEC,
        _process_instance_arguments,
    ),
]



"""Raw text export of ESL requirement nodes."""
from textwrap import indent
from typing import Any, Dict, List, Optional

from raesl.doc import lines
from ragraph.graph import Graph
from ragraph.node import Node


def requirement_text(requirement: Node, graph: Graph, skip: Optional[str] = "world") -> str:
    """Re-format the requirement as text."""
    info = requirement.annotations.esl_info
    if requirement.kind == "function_spec":
        if info["sub_kind"] == "goal":
            func = goal_text
        elif info["sub_kind"] == "transformation":
            func = transformation_text
        else:
            raise ValueError(f"Unknown sub kind for function spec: {info['sub_kind']}.")
    elif requirement.kind == "design_spec":
        func = design_text
    elif requirement.kind == "behavior_spec":
        func = behavior_text
    elif requirement.kind == "need":
        func = need_text
    else:
        raise ValueError(f"Unknown requirement type of node {requirement.json_dict}.")

    text = func(requirement, graph, skip=skip)

    if (
        requirement.kind == "function_spec"
        and info["body"].get("subclauses")
        or requirement.kind == "design_spec"
        and info["sub_clauses"]
    ):
        return "{}, with subclauses:\n{}".format(
            text, subclauses_text(requirement, graph, skip=skip, spaces=2)
        )
    else:
        return text


def goal_text(requirement: Node, graph: Graph, skip: Optional[str] = "world") -> str:
    """Re-format goal requirement as text."""
    body = requirement.annotations.esl_info["body"]
    text = "{active} {auxiliary} {verb} {variables} {preposition} {passive}".format(
        active=lines.node_path(body["active"], italic=False, arrows=False, skip=skip),
        auxiliary=body["auxiliary"],
        verb=body["verb"],
        variables=", ".join(
            lines.var_path(graph[var], italic=False, arrows=False, skip=skip)
            for var in body["variables"]
        ),
        preposition=body["preposition"],
        passive=lines.node_path(body["passive"], italic=False, arrows=False, skip=skip),
    )
    return text


def transformation_text(requirement: Node, graph: Graph, skip: Optional[str] = "world") -> str:
    """Re-format transformation requirement as text."""
    body = requirement.annotations.esl_info["body"]
    text = ("{auxiliary} {verb} {input_variables} {preposition} {output_variables}").format(
        auxiliary=body["auxiliary"],
        verb=body["verb"],
        input_variables=", ".join(
            lines.var_path(graph[var], italic=False, arrows=False, skip=skip)
            for var in body["input_variables"]
        ),
        preposition=body["preposition"],
        output_variables=", ".join(
            lines.var_path(graph[var], italic=False, arrows=False, skip=skip)
            for var in body["output_variables"]
        ),
    )
    return text


def designrule_text(
    body: Dict[str, Any],
    graph: Graph,
    skip: Optional[str] = "world",
) -> str:
    """Re-format design rule as text."""
    return "{subject} {auxiliary} {comparison} {value} {unit}".format(
        subject=lines.var_path(graph[body["subject"]], italic=False, arrows=False, skip=skip),
        auxiliary=body["auxiliary"],
        comparison=body["comparison"],
        value=body["bound"]["value"],
        unit=body["bound"]["unit"],
    )


def designclause_text(
    bodies: List[Dict[str, Any]],
    graph: Graph,
    label: Optional[str] = None,
    skip: Optional[str] = "world",
) -> str:
    """Re-format design clause as text."""
    rule = " or ".join(designrule_text(rule, graph, skip=skip) for rule in bodies)
    return "{}: {}".format(label, rule) if label else rule


def design_text(requirement: Node, graph: Graph, skip: Optional[str] = "world") -> str:
    """Re-format design requirement as text."""
    return designclause_text(requirement.annotations.esl_info["body"], graph, skip=skip)


def subclauses_text(
    requirement: Node, graph: Graph, skip: Optional[str] = "world", spaces: int = 2
) -> str:
    """Re-format subclauses as text."""
    if requirement.kind == "function_spec":
        clauses = requirement.annotations.esl_info["body"]["subclauses"]
    elif requirement.kind == "design_spec":
        clauses = requirement.annotations.esl_info["sub_clauses"]
    else:
        raise ValueError("Unsupported node kind '{}'.".format(requirement.kind))
    return "\n".join(
        indent(
            designclause_text(clause["body"], graph, label=clause["name"], skip=skip),
            prefix="{}{}".format(spaces * " ", "- " if spaces else ""),
        )
        for clause in clauses
    )


def case_text(case: Dict[str, Any], graph: Graph, skip: str = "world") -> str:
    """Re-format behavior requirement case as text."""
    return "{name}:\n  when:\n{whens}\n  then:\n{thens}".format(
        name=case["name"],
        whens=indent(
            "\n".join(
                designclause_text(clause["body"], graph, label=clause["name"], skip=skip)
                for clause in case["when_clauses"]
            ),
            prefix=4 * " ",
        ),
        thens=indent(
            "\n".join(
                designclause_text(clause["body"], graph, label=clause["name"], skip=skip)
                for clause in case["then_clauses"]
            ),
            prefix=4 * " ",
        ),
    )


def behavior_text(requirement: Node, graph: Graph, skip: Optional[str] = "world") -> str:
    """Re-format behavior requirement as text."""
    info = requirement.annotations.esl_info

    text = "\n\n".join(case_text(case, graph, skip=skip) for case in info["cases"])
    if info["default"]:
        text += "\n\nwhen no other case applies:\n{}".format(
            indent(
                "\n".join(
                    designclause_text(clause["body"], graph, label=clause["name"], skip=skip)
                    for clause in info["default"]
                ),
                prefix=4 * " ",
            ),
        )
    return text


def need_text(requirement: Node, graph: Graph, skip: Optional[str] = "world") -> str:
    """Re-format need as text."""
    info = requirement.annotations.esl_info
    text = ("{subject} {text}").format(
        subject=lines.var_path(graph[info["subject"]], italic=False, arrows=False, skip=skip),
        text=info["text"],
    )
    return text



"""ESL compiler Command Line Interface."""
import sys
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import click
from raesl import logger
from raesl.utils import check_output_path, get_esl_paths

# Compile module is excluded during docs generation.
try:
    from raesl.compile import parser, scanner
    from raesl.compile.instantiating import graph_building
except ImportError:
    pass

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.diagnostics import DiagnosticStore
    from ragraph.graph import Graph


def run(
    *paths: Union[str, Path],
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
    files: Optional[Union[List[str], List[Path]]] = None,
) -> Tuple["DiagnosticStore", Optional["Specification"], Optional["Graph"]]:
    """Run the compiler on ESL files.

    Arguments:
        paths: Paths to resolve into ESL files. May be any number of files and
            directories to scan.
        output: Optional output file (JSON) to write the graph to.
        force: Whether to overwrite the output file or raise an error if the file
            already exists.
        files: Optional paths argument (deprecated).

    Returns:
        Diagnostic storage.
        Specification object (if successfully parsed).
        Instantiated graph (if successfully instantiated).
    """
    if files is not None:
        paths = tuple(files)
        msg = " ".join(
            (
                "The 'files' keyword argument will be deprecated.",
                "Please use your file and directory paths as (any number of)",
                "positional arguments to this function.",
                "Also, take a look at 'raesl.compile.to_graph'",
                "or its alias 'ragraph.io.esl.from_esl'. to obtain a Graph.",
            )
        )
        logger.warning(msg)

    try:
        in_files = get_esl_paths(*paths)
        out_file = None if output is None else check_output_path(output, force)
    except ValueError as e:
        if click.get_current_context(silent=True):
            logger.error(str(e))
            sys.exit(1)
        raise e

    # Parse lexers per file.
    diag_store, spec = parser.parse_spec(
        scanner.Lexer(str(f), f.read_text(), 0, 0, 0, []) for f in in_files
    )

    # Errors have been reported to stdout already.
    if diag_store.has_severe() or spec is None:
        if click.get_current_context(silent=True):
            sys.exit(1)
        return diag_store, None, None

    # Succeeded so far: instantiate.
    graph = graph_building.GraphFactory(diag_store=diag_store, spec=spec).make_graph()
    if diag_store.has_severe() or graph is None:
        if click.get_current_context(silent=True):
            sys.exit(1)
        return diag_store, spec, None

    if output is not None:
        from ragraph.io.json import to_json

        to_json(graph, path=out_file)

    return diag_store, spec, graph


@click.command("compile")
@click.argument("paths", nargs=-1, type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option(
    "--output",
    "-o",
    default=None,
    type=click.Path(file_okay=True, dir_okay=False),
    help="Graph output file.",
)
@click.option(
    "--force",
    "-f",
    default=False,
    is_flag=True,
    help="Whether to overwrite an existing output file.",
)
def compile(paths: List[str], output: Optional[str], force: bool):
    """Run the ESL compiler."""
    run(*paths, output=output, force=force)



"""Line matcher state machines for transformations."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_DEFINE_COMPONENT_SPEC = """
define_component:
    start initial;
    start -> s1 [DEFINE_KW];
        s1 -> s2 [COMPONENT_KW];
        s2 -> s3 [NAME] tag=comp_name;
    start -> s3 [WORLD_KW] tag=world;

    end accept=define_component;
    s3 -> end [NL_TK];
    s3 -> end [EOF_TK];
"""


def _process_compdef(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    if "world" in tags:
        pos_tok = tags["world"][0]
        name_tok = None
    else:
        name_tok = tags["comp_name"][0]
        pos_tok = name_tok

    builder.notify_new_section(pos_tok, True)
    builder.compdef_builder.new_componentdef(pos_tok, name_tok)


_EMPYTY_SPEC = """
empty_component:
    start initial;
    start -> s1 [EMPTY_KW];

    end accept=empty_component;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""

# No processing function needed for _EMPYTY_SPEC lines.

MACHINES: typing.MachineTripletList = [
    ("DEFINE_COMPONENT_MACHINE", _DEFINE_COMPONENT_SPEC, _process_compdef),
    ("EMPTY_MACHINE", _EMPYTY_SPEC, None),
]



"""Builder to add child component instances to a component definition."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple

from raesl.compile import diagnostics
from raesl.compile.ast.components import ComponentInstance, InstanceArgument
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.type_checker import check_type

if TYPE_CHECKING:
    from raesl.compile.ast.components import ComponentDefinition
    from raesl.compile.ast.nodes import Node
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CompDefCompInstBuilder:
    """Collect and check child component instances of a component definition.

    Arguments:
        comp_child_builders: Storage of child builder for a component definition.

    Attributes:
        diag_store: Child builders problem store.
        instances: Collected component instances.
        last_instance: Link to last added instance, to allow adding instance
            arguments to it.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.instances: List[ComponentInstance] = []
        self.comp_child_builders = comp_child_builders
        self.last_instance: Optional[ComponentInstance] = None

    def add_compinst(self, inst_name_tok: "Token", def_name_tok: "Token", has_arguments: bool):
        """Store a new child component instance line."""
        compinst = ComponentInstance(inst_name_tok, def_name_tok)
        self.instances.append(compinst)
        if has_arguments:
            self.last_instance = compinst
        else:
            self.last_instance = None

    def add_compinst_arguments(self, arguments: List["Token"]):
        """Store a line of component instance argument names."""
        assert self.last_instance is not None
        for argname in arguments:
            self.last_instance.arguments.append(InstanceArgument(argname))

    def get_compdef_names(self) -> Set[str]:
        """Get the names of used definitions."""
        def_names = set()
        for inst in self.instances:
            name = inst.def_name_tok.tok_text
            def_names.add(name)

        return def_names

    def finish_comp(self, comp_def: "ComponentDefinition", spec: "Specification"):
        """Finish checking and adding child component instances to the component.

        Arguments:
            comp_def: Used as 'my' component definition.
            spec: Used as source for types.
        """
        # Available names of variables, parameters, and variable-groups in 'my'
        # component.
        avail_vps = utils.construct_var_param_map(comp_def)
        avail_vgroups = utils.construct_vargroup_map(comp_def)

        # The 'other' component definition needed for checking the instance  is
        # already available, as ComponentDefBuilder ordered compdef checking on
        # instance use.
        avail_compdefs = dict(
            (cdef.name_tok.tok_text, cdef) for cdef in spec.comp_defs if cdef.name_tok is not None
        )

        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for inst in self.instances:
            elements_by_label[inst.inst_name_tok.tok_text].append(inst)

        # Collection reported names to avoid double problem reports.
        reported_names: Set[str] = set()

        # Check 'my' instances.
        for inst in self.instances:
            compdef = avail_compdefs.get(inst.def_name_tok.tok_text)
            if compdef is None:
                # Truly undefined 'other' component definition.
                name = inst.def_name_tok.tok_text
                loc = inst.def_name_tok.get_location()
                self.diag_store.add(
                    diagnostics.E203("component definition", name=name, location=loc)
                )
                continue  # Cannot do anything else useful with it.

            # Link 'other' component definition to the instance for future use.
            inst.compdef = compdef

            # Found a component definition with the same name, do arguments match as
            # well?
            #
            # 1. Collect argument information of the instance.
            found_error = False
            inst_arguments: List[Optional[Tuple["Token", "Node"]]] = []
            for arg in inst.arguments:
                node: Optional["Node"]
                node = utils.resolve_var_param_group_node(
                    arg.name_tok,
                    avail_vps,
                    avail_vgroups,
                    reported_names,
                    self.diag_store,
                )

                if node is None:
                    # resolve_var_param_group_node() already created an error.

                    inst_arguments.append(None)  # Add dummy to check other arguments.
                    found_error = True
                else:
                    inst_arguments.append((arg.name_tok, node))
                    arg.argnode = node

            # Convert parameters of the definition to InputType as well.
            parameters = [(param.name_tok, param.node) for param in compdef.parameters]

            # 2. Check number of parameters against number of arguments.
            inst_length = len(inst_arguments)
            def_length = len(parameters)
            if inst_length != def_length:
                self.diag_store.add(
                    diagnostics.E221(
                        "argument",
                        inst_length,
                        def_length,
                        location=inst.inst_name_tok.get_location(),
                        references=[compdef.pos_tok.get_location()],
                    )
                )
                found_error = True
                continue

            # 3. Check argument types against parameter types of compdef, skipping None
            #    arguments. We assume undirected flow direction, thus arguments must
            #    accept all possible values that the definition may provide.
            for param_input, arg_input in zip(parameters, inst_arguments):
                if arg_input is None:
                    continue  # Already reported as an error.

                diag = check_type(
                    subtype=arg_input, supertype=param_input, allow_subtype_limits=False
                )
                if diag:
                    self.diag_store.add(diag)
                    found_error = True

            # Add inst to compdef if no errors.
            if not found_error:
                comp_def.component_instances.append(inst)



"""ESL compiler Diagnostics.

Diagnostic code scheme:

    Severity:

        E###: ERROR
        W###: WARNING
        I###: INFO
        H###: HINT


    Origin:

        #000: General
        #100: Scanning/Parsing
        #200: Typechecking
        #300: AST Builder
        #400: Instance/output builder
"""

import sys
from typing import IO, Iterable, List, Optional

import click

from raesl import logger, utils
from raesl.types import (
    Diagnostic,
    DiagnosticRelatedInformation,
    DiagnosticSeverity,
    Location,
)

# Proxies
ERROR = DiagnosticSeverity.Error
WARN = DiagnosticSeverity.Warning
INFO = DiagnosticSeverity.Information
HINT = DiagnosticSeverity.Hint
_NON_SEVERE = {WARN, INFO, HINT}
EslRelated = DiagnosticRelatedInformation


class EslDiagnostic(Diagnostic):
    """An unscoped diagnostic as ESL works with multiple text documents at once.

    Arguments:
        message: The diagnostic's message.
        location: The location at which the message applies.
        severity: The diagnostic's severity. Can be omitted. If omitted it is up to the
            client to interpret diagnostics as error, warning, info or hint.
        code: The diagnostic's code, which might appear in the user interface.
        source: A human-readable string describing the source of this diagnostic, e.g.
            'esl', or 'esl compiler'.
        related_information: A list of related diagnostic information, e.g. when
            symbol-names within a scope collide you can mark all definitions via this
            property.

    Attributes:
        range: The range at which the message applies.
    """

    def __init__(
        self,
        message: str,
        location: Location = utils.get_location(),
        severity: DiagnosticSeverity = ERROR,
        code: str = "E100",
        source: str = "RaESL compiler",
        related_information: List[DiagnosticRelatedInformation] = None,
    ):
        self.location = location
        range = None if location is None else location.range
        super().__init__(message, range, severity, code, source, related_information)

    def __repr__(self):
        # Debug/internal: 0-based.
        return self._print(style=repr)

    def __str__(self):
        # Display: 1-based.
        return self._print(style=str)

    def _print(self, style=repr):
        delim = "\n    "
        headline = (
            f"{self.severity.name.upper()} {self.source} [{self.code}]"
            + f" at {style(self.location)}\n  {self.message}"
        )
        sublines = (
            delim.join(style(i) for i in self.related_information)
            if self.related_information is not None
            else None
        )
        if sublines is None:
            return headline
        else:
            return headline + delim + sublines


class DiagnosticStore:
    """Storage of found diagnostics.

    Attributes:
        diagnostics: Stored diagnostics.
        severity: What diagnostics to log.
        exit: Exit on error.
    """

    def __init__(self, severity: DiagnosticSeverity = ERROR, exit: bool = False):
        self.diagnostics: List[EslDiagnostic] = []
        self.severity = severity
        self.exit = exit

    def add(self, diagnostic: EslDiagnostic):
        """Add a diagnostic. Report directly if below severity threshold."""
        self.diagnostics.append(diagnostic)
        if diagnostic.severity <= self.severity:
            self.report(diagnostic)

    def report(self, diagnostic: EslDiagnostic):
        """Report a single diagnostic."""
        message = str(diagnostic)  # Human readable (e.g, 1-based instead of 0-based)

        if diagnostic.severity <= ERROR:
            logger.error(message)
            if self.exit:
                sys.exit(1)
        elif diagnostic.severity <= WARN:
            logger.warning(message)
        else:
            logger.info(message)

    def dump(self, test: bool = False, stream: Optional[IO[str]] = None):
        """Dump all stored diagnostics to the given stream.

        Arguments:
            test: Whether to output in test mode (otherwise: user-friendly).
            stream: Output stream to use. Defaults to stdout.
        """
        stream = click.get_text_stream("stdout") if stream is None else stream

        if not test and not self.diagnostics:
            stream.write("No diagnostics to report.")

        for d in self.diagnostics:
            stream.write(str(d))
            stream.write("\n")

    def has_severe(self) -> bool:
        """Whether there are severe diagnostics stored."""
        return any(d.severity not in _NON_SEVERE for d in self.diagnostics)


def E100(location: Location = utils.get_location()) -> EslDiagnostic:
    """Unexpected end of the specification."""
    return EslDiagnostic(
        "Unexpected end of the specification.",
        location=location,
        severity=ERROR,
        code="E100",
        source="ESL parser",
    )


def E101(acceptors: Iterable[str], location: Location = utils.get_location()) -> EslDiagnostic:
    """Best line match is ambiguous. Found multiple acceptors: {ambi_acceptors}. This
    is an internal error.
    """
    enum = "', '".join(acceptors)
    return EslDiagnostic(
        (
            f"Best line match is ambiguous. Found multiple acceptors: '{enum}'. "
            + "This is most likely an internal error."
        ),
        location=location,
        severity=ERROR,
        code="E101",
        source="ESL parser",
    )


def E102(location: Location = utils.get_location()) -> EslDiagnostic:
    """Syntax error."""
    return EslDiagnostic(
        "Syntax error.",
        location=location,
        severity=ERROR,
        code="E102",
        source="ESL parser",
    )


def E200(
    name: str,
    kind: str,
    location: Location = utils.get_location(),
    dupes: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Multiple {kind} named '{name}'."""
    dupes = [] if dupes is None else dupes
    return EslDiagnostic(
        f"Multiple {kind}s named '{name}'.",
        location=location,
        severity=ERROR,
        code="E200",
        source="ESL typechecker",
        related_information=[EslRelated(dupe, f"Duplicate {kind} '{name}'.") for dupe in dupes],
    )


def E201(section: str, context: str, location: Location = utils.get_location()) -> EslDiagnostic:
    """This '{section}' section is not allowed in the '{context}' context."""
    return EslDiagnostic(
        f"This '{section}' section is not allowed in the '{context}' context.",
        location=location,
        severity=ERROR,
        code="E201",
        source="ESL typechecker",
    )


def E202(
    kind: str,
    name: Optional[str] = None,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """Missing {kind} for '{name}'."""
    if name is None:
        msg = f"Missing {kind}."
    else:
        msg = f"Missing {kind} for '{name}'."
    return EslDiagnostic(
        msg,
        location=location,
        severity=ERROR,
        code="E202",
        source="ESL typechecker",
    )


def E203(
    kind: str, name: Optional[str] = None, location: Location = utils.get_location()
) -> EslDiagnostic:
    """Unknown {kind} named '{name}'."""
    if name is None:
        msg = f"Unknown {kind}."
    else:
        msg = f"Unknown {kind} named '{name}'."
    return EslDiagnostic(
        msg,
        location=location,
        severity=ERROR,
        code="E203",
        source="ESL typechecker",
    )


def E204(
    name: str,
    kind: str,
    location: Location = utils.get_location(),
    cycle: Optional[List[Location]] = None,
):
    """Cyclically dependent {kind} named '{name}'."""
    cycle = [] if cycle is None else cycle
    length = len(cycle)
    return EslDiagnostic(
        f"Cyclically dependent {kind} named '{name}'.",
        location=location,
        severity=ERROR,
        code="E204",
        source="ESL typechecker",
        related_information=[
            EslRelated(entry, f"Cycle {i+1}/{length}.") for i, entry in enumerate(cycle)
        ],
    )


def E205(name: str, context: str, location: Location = utils.get_location()) -> EslDiagnostic:
    """Cannot find {name} in {context}."""
    return EslDiagnostic(
        f"Cannot find {name} in {context}.",
        location=location,
        severity=ERROR,
        code="E205",
        source="ESL typechecker",
    )


def E206(
    name: str,
    kind: str,
    location: Location = utils.get_location(),
    blocks: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Found {kind} block(s), but the relation definition {name} has no such
    parameters.
    """
    blocks = [] if blocks is None else blocks
    return EslDiagnostic(
        (
            f"Found {kind} block(s), "
            + f"but the relation definition {name} has no such parameters."
        ),
        location=location,
        severity=ERROR,
        code="E206",
        source="ESL typechecker",
        related_information=[EslRelated(block, f"Relation {kind} block") for block in blocks],
    )


def E207(
    name: str,
    kind: str,
    location: Location = utils.get_location(),
    definition: Location = utils.get_location(),
) -> EslDiagnostic:
    """Relation instance '{name}' is missing a '{kind}' parameters section."""
    return EslDiagnostic(
        f"Relation instance '{name}' is missing a '{kind}' parameters section.",
        location=location,
        severity=ERROR,
        code="E207",
        source="ESL typechecker",
        related_information=[EslRelated(definition, "Corresponding relation definition.")],
    )


def E208(
    name: str,
    kind: str,
    num: int,
    location: Location = utils.get_location(),
    definition: Location = utils.get_location(),
) -> EslDiagnostic:
    """Relation instance '{name}' is missing at least {num} '{kind}' parameters."""
    return EslDiagnostic(
        f"Relation instance '{name}' is missing at least {num} '{kind}' parameters.",
        location=location,
        severity=ERROR,
        code="E208",
        source="ESL typechecker",
        related_information=[EslRelated(definition, "Corresponding relation definition.")],
    )


def E209(
    name: str,
    kind: str,
    other_kind: str,
    location: Location = utils.get_location(),
    others: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """'{name}' is both a {kind} and a {other_kind}."""
    others = [] if others is None else others
    return EslDiagnostic(
        f"'{name}' is both a {kind} and a {other_kind}.",
        location=location,
        severity=ERROR,
        code="E209",
        source="ESL typechecker",
        related_information=[EslRelated(other, f"{other_kind} location.") for other in others],
    )


def E210(lhs: Location, rhs: Location, reason: str = "are not compatible") -> EslDiagnostic:
    """Values cannot be compared, they {reason}."""
    return EslDiagnostic(
        f"Values cannot be compared, they {reason}.",
        location=lhs,
        severity=ERROR,
        code="E210",
        source="ESL typechecker",
        related_information=[EslRelated(rhs, "Other value location.")],
    )


def E211(
    verb: str,
    preposition: str,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """Unsupported verb-preposition combination '{verb} {preposition}'."""
    return EslDiagnostic(
        f"Unsupported verb-preposition combination '{verb} {preposition}'.",
        location=location,
        severity=ERROR,
        code="E211",
        source="ESL typechecker",
    )


def E212(
    kind: str,
    value: str,
    allowed: str,
    name: Optional[str] = None,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """{kind.capitalize()} '{name}' uses '{value}', but should use {allowed}."""
    if name is None:
        msg = f"{kind.capitalize()} uses '{value}', but should use {allowed}."
    else:
        msg = f"{kind.capitalize()} '{name}' uses '{value}', but should use {allowed}."
    return EslDiagnostic(
        msg,
        location=location,
        severity=ERROR,
        code="E212",
        source="ESL typechecker",
    )


def E213(
    kind: str,
    num: int,
    allowed: str,
    location: Location = utils.get_location(),
    occurrences: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Found {num} {kind}(s), but there should be {allowed}."""
    occurrences = [] if occurrences is None else occurrences
    return EslDiagnostic(
        f"Found {num} {kind}(s), but there should be {allowed}.",
        location=location,
        severity=ERROR,
        code="E213",
        source="ESL typechecker",
        related_information=[EslRelated(occ, f"{kind.capitalize()}") for occ in occurrences],
    )


def E214(
    name: str,
    location: Location = utils.get_location(),
    def_location: Optional[Location] = None,
) -> EslDiagnostic:
    """Definition of type '{name}' failed with an error."""
    def_locs = [] if def_location is None else [def_location]
    return EslDiagnostic(
        f"Definition of type '{name}' failed with an error.",
        location=location,
        severity=ERROR,
        code="E214",
        source="ESL typechecker",
        related_information=[EslRelated(loc, "Related type definition.") for loc in def_locs],
    )


def E215(
    name: str,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """Type name '{name}' must be the name of an elementary type."""
    return EslDiagnostic(
        f"Type name '{name}' must be the name of an elementary type.",
        location=location,
        severity=ERROR,
        code="E215",
        source="ESL typechecker",
    )


def E216(
    name: str,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """Unit '{name}' should not have square brackets around it's name."""
    return EslDiagnostic(
        f"Unit '{name}' should not have square brackets around it's name.",
        location=location,
        severity=ERROR,
        code="E216",
        source="ESL typechecker",
    )


def E217(
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """The dimensionless unit '-' is not allowed to be specified explicitly."""
    return EslDiagnostic(
        "The dimensionless unit '-' is not allowed to be specified explicitly.",
        location=location,
        severity=ERROR,
        code="E217",
        source="ESL typechecker",
    )


def E218(name: str, location: Location = utils.get_location()) -> EslDiagnostic:
    """Standard type '{name}' cannot be overridden."""
    return EslDiagnostic(
        f"Standard type '{name}' cannot be overridden.",
        location=location,
        severity=ERROR,
        code="E218",
        source="ESL typechecker",
    )


def E219(name: str, location: Location = utils.get_location()) -> EslDiagnostic:
    """Unit '{name}' is not allowed here."""
    return EslDiagnostic(
        f"Unit '{name}' is not allowed here.",
        location=location,
        severity=ERROR,
        code="E219",
        source="ESL typechecker",
    )


def E220(name: str, kind: str, location: Location = utils.get_location()) -> EslDiagnostic:
    """Element '{name}' does not match with a {kind}."""
    return EslDiagnostic(
        f"Element '{name}' does not match with a {kind}.",
        location=location,
        severity=ERROR,
        code="E220",
        source="ESL typechecker",
    )


def E221(
    kind: str,
    num: int,
    expected: int,
    location: Location = utils.get_location(),
    references: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Number of {kind}s does not match. Found {num}, expected {expected}."""
    references = [] if references is None else references
    return EslDiagnostic(
        f"Number of {kind}s does not match. Found {num}, expected {expected}.",
        location=location,
        severity=ERROR,
        code="E221",
        source="ESL typechecker",
        related_information=[
            EslRelated(ref, f"Reference with {expected} {kind}(s).") for ref in references
        ],
    )


def E222(
    name: str,
    other: str,
    location: Location = utils.get_location(),
    other_loc: Location = utils.get_location(),
) -> EslDiagnostic:
    """Value '{name}' has additional value restrictions relative to '{other}'."""
    return EslDiagnostic(
        f"Value '{name}' has additional value restrictions relative to '{other}'.",
        location=location,
        severity=ERROR,
        code="E222",
        source="ESL typechecker",
        related_information=[EslRelated(other_loc, f"Other value '{other}'.")],
    )


def E223(
    name: str,
    other: str,
    kind: str,
    location: Location = utils.get_location(),
    other_loc: Optional[Location] = None,
) -> EslDiagnostic:
    """'{name}' is not a {kind} of {other}."""
    related = [] if other_loc is None else [EslRelated(other_loc, f"{other}")]
    return EslDiagnostic(
        f"'{name}' is not a {kind} of {other}.",
        location=location,
        severity=ERROR,
        code="E223",
        source="ESL typechecker",
        related_information=related,
    )


def E224(
    kind: str,
    unsupported: str,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """{kind.capitalize()}s do not support {unsupported}."""
    return EslDiagnostic(
        f"{kind.capitalize()}s do not support {unsupported}.",
        location=location,
        severity=ERROR,
        code="E224",
        source="ESL typechecker",
    )


def E225(
    part: str,
    first_part: str,
    kind: str,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """Cannot resolve '.{part}' part of the '{first_part}' {kind}."""
    return EslDiagnostic(
        f"Cannot resolve '.{part}' part of the '{first_part}' {kind}.",
        location=location,
        severity=ERROR,
        code="E225",
        source="ESL typechecker",
    )


def E226(
    name: str,
    location: Location = utils.get_location(),
) -> EslDiagnostic:
    """Need '{name}' is not allowed to reference a bundle."""
    return EslDiagnostic(
        f"Need '{name}' is not allowed to reference a bundle.",
        location=location,
        severity=ERROR,
        code="E226",
        source="ESL typechecker",
        related_information=[EslRelated(location, "Only elementary variables, try its fields.")],
    )


def E227(
    name: str,
    scope: str,
    location: Location = utils.get_location(),
    dupes: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Multiple identifier '{name}' within '{cdef_name}'."""
    dupes = [] if dupes is None else dupes
    return EslDiagnostic(
        f"Duplicate identifier '{name}' within the scope of '{scope}'.",
        location=location,
        severity=ERROR,
        code="E200",
        source="ESL typechecker",
        related_information=[EslRelated(dupe, f"Duplicate identifier '{name}'.") for dupe in dupes],
    )


def E400(
    name: str,
    location: Location = utils.get_location(),
    owners: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Elementary variable value '{name}' has more than one property owner."""
    owners = [] if owners is None else owners
    return EslDiagnostic(
        f"Elementary variable value '{name}' has more than one property owner.",
        location=location,
        severity=ERROR,
        code="E400",
        source="ESL instantiating",
        related_information=[EslRelated(owner, "Duplicate owner.") for owner in owners],
    )


def W200(
    name: str,
    kind: str,
    location: Location = utils.get_location(),
    dupes: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """{kind.capitalize()} '{name}' has been specified multiple times."""
    dupes = [] if dupes is None else dupes
    return EslDiagnostic(
        f"{kind.capitalize()} '{name}' has been specified multiple times.",
        location=location,
        severity=WARN,
        code="W200",
        source="ESL typechecker",
        related_information=[EslRelated(dupe, f"Duplicate {kind}.") for dupe in dupes],
    )


def W300(
    element: Optional[str] = None,
    location: Location = utils.get_location(),
    comments: Optional[List[Location]] = None,
) -> EslDiagnostic:
    """Documentation comment(s) could not be assigned to '{element}'."""
    element = "a specification element" if element is None else f"'{element}'"
    comments = [] if comments is None else comments
    return EslDiagnostic(
        f"Documentation comment(s) could not be assigned to {element}.",
        location=location,
        severity=WARN,
        code="W300",
        source="ESL AST builder",
        related_information=[
            EslRelated(doc, "Unassigned documentation comment.") for doc in comments
        ],
    )



"""Some general types, inspired by the Language Server Protocol.

Reference:
    https://microsoft.github.io/language-server-protocol/
"""
import enum
from pathlib import Path
from typing import List, Tuple


class Position:
    """Position in a text document expressed as zero-based line and character offset.

    A position is between two characters like an 'insert' cursor in an editor. Special
    values like for example -1 to denote the end of a line are not supported.

    Arguments:
        line: Line position in a document (zero-based).
        character: Character offset on a line in a document (zero-based). Assuming that
            the line is represented as a string, the 'character' value represents the
            gap between 'character' and 'character + 1'.
    """

    def __init__(self, line: int = 0, character: int = 0):
        self.line = line
        self.character = character

    def __eq__(self, other):
        return (
            isinstance(other, Position)
            and self.line == other.line
            and self.character == other.character
        )

    def __ge__(self, other):
        line_gt = self.line > other.line

        if line_gt:
            return line_gt

        if self.line == other.line:
            return self.character >= other.character

        return False

    def __gt__(self, other):
        line_gt = self.line > other.line

        if line_gt:
            return line_gt

        if self.line == other.line:
            return self.character > other.character

        return False

    def __le__(self, other):
        line_lt = self.line < other.line

        if line_lt:
            return line_lt

        if self.line == other.line:
            return self.character <= other.character

        return False

    def __lt__(self, other):
        line_lt = self.line < other.line

        if line_lt:
            return line_lt

        if self.line == other.line:
            return self.character < other.character

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.line, self.character))

    def __iter__(self):
        return iter((self.line, self.character))

    def __repr__(self):
        # Debug/internal: 0-based.
        return f"{self.line}:{self.character}"

    def __str__(self):
        # Display: 1-based.
        return f"{self.line + 1}:{self.character + 1}"


class Range:
    """A range in a text document expressed as (zero-based) start and end positions.

    A range is comparable to a selection in an editor. Therefore the end position is
    exclusive. If you want to specify a range that contains a line including the ending
    character(s) then use an end position denoting the start of the next line.

    Arguments:
        start: The range's start position.
        end: The range's end position.
    """

    def __init__(self, start: Position, end: Position):
        self.start = start
        self.end = end

    def __eq__(self, other):
        return isinstance(other, Range) and self.start == other.start and self.end == other.end

    def __hash__(self):
        return hash((self.start, self.end))

    def __iter__(self):
        return iter((self.start, self.end))

    def __repr__(self):
        # Debug/internal: 0-based.
        return f"{repr(self.start)}-{repr(self.end)}"

    def __str__(self):
        # Display: 1-based.
        return f"{str(self.start)}-{str(self.end)}"


class Location:
    """Represents a location inside a resource. Such as a line inside a text file.

    Arguments:
        uri: URI of this location.
        range: Range of this location.
    """

    def __init__(self, uri: str, range: Range):
        self.uri = uri
        self.range = range

    def get_key(self) -> Tuple[str, int, int]:
        """Get a tuple with identification for this position."""
        return self.uri, self.range.start.line, self.range.start.character

    def __eq__(self, other):
        return isinstance(other, Location) and self.uri == other.uri and self.range == other.range

    def __lt__(self, other: "Location"):
        return self.get_key() < other.get_key()

    def __le__(self, other: "Location"):
        return self.get_key() <= other.get_key()

    def __gt__(self, other: "Location"):
        return self.get_key() > other.get_key()

    def __ge__(self, other: "Location"):
        return self.get_key() >= other.get_key()

    def __repr__(self):
        # Debug/internal: 0-based.
        return f"{Path(self.uri).as_posix()}:{repr(self.range)}"

    def __str__(self):
        # Display: 1-based.
        return f"{Path(self.uri).as_posix()}:{str(self.range)}"


class DiagnosticSeverity(enum.IntEnum):
    Error = 1  # Reports an error.
    Warning = 2  # Reports a warning.
    Information = 3  # Reports an information.
    Hint = 4  # Reports a hint.


class DiagnosticRelatedInformation:
    """Represents a related message and source code location for a diagnostic.

    This should be used to point to code locations that cause or are related to a
    diagnostic, e.g when duplicating a symbol in a scope.

    Arguments:
        location: The location of this related diagnostic information.
        message: The message of this related diagnostic information.
    """

    def __init__(self, location: Location, message: str):
        self.location = location
        self.message = message

    def __repr__(self):
        return f"{self.message} at {repr(self.location)}"

    def __str__(self):
        return f"{self.message} at {str(self.location)}"


class Diagnostic:
    """Represents a diagnostic, such as a compiler error or warning.

    Diagnostic objects are only valid in the scope of a resource.

    Arguments:
        message: The diagnostic's message.
        range: The range at which the message applies.
        severity: The diagnostic's severity. Can be omitted. If omitted it is up to the
            client to interpret diagnostics as error, warning, info or hint.
        code: The diagnostic's code, which might appear in the user interface.
        source: A human-readable string describing the source of this diagnostic, e.g.
            'esl', or 'esl compiler'.
        related_information: A list of related diagnostic information, e.g. when
            symbol-names within a scope collide you can mark all definitions via this
            property.
    """

    def __init__(
        self,
        message: str,
        range: Range,
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
        code: str = None,
        source: str = None,
        related_information: List[DiagnosticRelatedInformation] = None,
    ):
        self.message = message
        self.range = range
        self.severity = severity
        self.code = code
        self.source = source
        self.related_information = related_information

    def __repr__(self):
        # Debug/internal: 0-based.
        return self._print(style=repr)

    def __str__(self):
        # Display: 1-based.
        return self._print(style=str)

    def _print(self, style=repr):
        delim = "\n    "
        headline = (
            f"{self.severity.name.upper()} {self.source} [{self.code}]"
            + f" at {style(self.range)}\n  {self.message}"
        )
        sublines = (
            delim.join(style(i) for i in self.related_information)
            if self.related_information is not None
            else None
        )
        if sublines is None:
            return headline
        else:
            return headline + delim + sublines



"""AST storage of types."""
from typing import TYPE_CHECKING, List, Optional, Set, Tuple

if TYPE_CHECKING:
    from raesl.compile.scanner import Token
    from raesl.compile.ast.exprs import Value


class BaseType:
    """Base class of a type."""

    def get_units(self) -> Optional[Set[str]]:
        """Retrieve the units that may be used with values of the type.

        Returns:
            Set of unit names, set(['-']) if it has no units specified, or None if the
                type doesn't support units.
        """
        raise NotImplementedError("Implement me in {}.".format(repr(self)))


class ElementaryType(BaseType):
    """A type of a singular value in ESL. The allowed values in a parent type always
    have priority over the allowed values in a child type.

    Arguments:
        parent: Parent elementary type if specified.
        units: Allowed units of the type, should not have square brackets around the
            text.
        intervals: Disjunction of allowed ranges of the type, pairs of (lowerbound,
            upperbound) where one of the bounds may be None. Constants and enumerations
            are expressed as intervals with the same lower and upper bound.
    """

    def __init__(
        self,
        parent: Optional["ElementaryType"],
        units: List["Token"],
        intervals: Optional[List[Tuple[Optional["Value"], Optional["Value"]]]],
    ):
        super(ElementaryType, self).__init__()
        self.parent = parent
        self.units = units
        self.intervals = intervals

        assert parent is None or isinstance(parent, ElementaryType)

    def get_units(self):
        avail_units = set()
        etp = self
        while etp is not None:
            avail_units.update((unit.tok_text for unit in etp.units))
            etp = etp.parent

        if not avail_units:
            avail_units.add("-")
        return avail_units


class TypeDef:
    """A named type.

    Arguments:
        name: Name of the type definition.
        the_type: Type associated with the name.
    """

    def __init__(self, name: "Token", the_type: BaseType):
        self.name = name
        self.type = the_type


class CompoundField:
    """A named field in a Compound.

    Arguments:
        name: Name of the compound field.
        the_type: Type of the compound field.
    """

    def __init__(self, name: "Token", the_type: BaseType):
        self.name = name
        self.type = the_type


class Compound(BaseType):
    """A collection of named typed values. Note that a Compound cannot have parents,
    units, or intervals.

    Arguments:
        fields: Fields of the compound.
    """

    def __init__(self, fields: List[CompoundField]):
        super(Compound, self).__init__()
        self.fields = fields

    def get_units(self):
        return None



"""ESL Jupyter Kernel implementation."""

import json
import pathlib
import tempfile
import urllib
from typing import Optional

from ipykernel import get_connection_file
from ipykernel.kernelbase import Kernel
from notebook import notebookapp
from raesl import __version__

# Compile module is excluded during docs generation.
try:
    from raesl.compile.cli import run
except ImportError:
    pass


class EslKernel(Kernel):
    implementation = "raesl.compile"
    implementation_version = __version__
    language = "esl"
    language_version = "2.0"
    language_info = {
        "name": "Elephant Specification Language",
        "mimetype": "text/x-esl",
        "file_extension": ".esl",
    }
    banner = "ESL Kernel - Support for the Elephant Specification Language."

    def do_execute(
        self, code, silent, store_history=True, user_expressions=None, allow_stdin=False
    ):
        stdout = []
        stderr = []

        with tempfile.TemporaryDirectory() as tmpdir:
            esl = pathlib.Path(tmpdir) / "notebook.esl"
            esl.write_text(code, encoding="utf-8")
            try:
                run(esl)
                stdout.append("Compiler ran succesfully.")
                try:
                    # notebook = self.get_notebook_path()
                    # if not notebook:
                    #     notebook = pathlib.Path("./notebook.esl")
                    esl_doc = pathlib.Path("./notebook.esl")
                    esl_doc.write_text(code)
                    stdout.append("ESL file saved as {}.".format(esl_doc.name))
                except Exception as exc:
                    stderr.append(str(exc))
            except Exception as exc:
                stderr.append(str(exc))

        if not silent:
            stdout_content = {"name": "stdout", "text": "\n".join(stdout)}
            self.send_response(self.iopub_socket, "stream", stdout_content)

        if stderr:
            stderr_content = {"name": "stderr", "text": "\n".join(stderr)}
            self.send_response(self.iopub_socket, "stream", stderr_content)

        return {
            "status": "ok",
            # The base class increments the execution count
            "execution_count": self.execution_count,
            "payload": [],
            "user_expressions": {},
        }

    def get_notebook_path(self) -> Optional[pathlib.Path]:
        connection_file = get_connection_file()
        kernel_id = connection_file.split("-", 1)[1].split(".")[0]

        for srv in notebookapp.list_running_servers():
            try:
                if srv["token"] == "" and not srv["password"]:
                    req = urllib.request.urlopen(srv["url"] + "api/sessions")
                else:
                    req = urllib.request.urlopen(srv["url"] + "api/sessions?token=" + srv["token"])
                sessions = json.load(req)
                for sess in sessions:
                    if sess["kernel"]["id"] == kernel_id:
                        return pathlib.Path(srv["notebook_dir"], sess["notebook"]["path"])
            except Exception:
                return None



"""Library with common parts of line matchers."""
from typing import List

from raesl.compile.machine_files.utils import make_loc_names


def get_auxiliary_verb_part(start_loc: str, end_1locs: List[str], tagname: str) -> str:
    """Part implementing the 'auxiliary-verb' rule.

    auxiliary-verb ::=
        "shall" | "must" | "should" | "could" | "won't"
    """
    assert len(end_1locs) == 1

    text = """\
        {start} -> {end} [SHALL_KW] tag={tagname};
        {start} -> {end} [MUST_KW] tag={tagname};
        {start} -> {end} [SHOULD_KW] tag={tagname};
        {start} -> {end} [COULD_KW] tag={tagname};
        {start} -> {end} [WONT_KW] tag={tagname};
    """
    return text.format(start=start_loc, end=end_1locs[0], tagname=tagname)


def get_does_auxiliary_verb_part(start_loc: str, end_1locs: List[str], tagname: str) -> str:
    """State machine part that implements:

    "does" | auxiliary-verb
    """
    assert len(end_1locs) == 1

    text = """\
        {start} -> {end} [DOES_KW] tag={tagname};
    """ + get_auxiliary_verb_part(
        start_loc, [end_1locs[0]], tagname
    )

    return text.format(start=start_loc, end=end_1locs[0], tagname=tagname)


def get_is_auxiliary_verb_part(
    start_loc: str, end_1locs: List[str], tagname: str, prefix: str = ""
) -> str:
    """State machine part implementing:

    "is" | auxiliary-verb "be"
    """
    assert len(end_1locs) == 1

    locs = {"start": start_loc, "end": end_1locs[0], "tagname": tagname}
    locs.update(make_loc_names(prefix, "isauxloc", 1))

    text = """\
        {isauxloc1};

        {start} -> {end} [IS_KW] tag={tagname};
        {isauxloc1} -> {end} [BE_KW];
    """ + get_auxiliary_verb_part(
        start_loc, [locs["isauxloc1"]], tagname
    )

    return text.format(**locs)


def get_argument_references_part(
    start_loc: str, end_1locs: List[str], tagname: str, prefix: str = ""
) -> str:
    """State machine part implementing the 'argument-references' rule.

    argument-references ::=
        argument-name { and-connector argument-name } \n

    and-connector ::=
        "and" | "," | "," "and"
    """
    assert len(end_1locs) == 1

    locs = {"start": start_loc, "end": end_1locs[0], "tagname": tagname}
    locs.update(make_loc_names(prefix, "argrefs", 1))

    text = """\
        {argrefs1};

        {start}    -> {end}      [DOTTEDNAME] tag={tagname};
        {end}      -> {start}    [AND_KW];
        {end}      -> {argrefs1} [COMMA_TK];
        {argrefs1} -> {end}      [DOTTEDNAME] tag={tagname};
        {argrefs1} -> {start}    [AND_KW];
    """

    return text.format(**locs)


def get_compare_op_part(
    start_loc: str, end_1locs: List[str], tagname: str, prefix: str = ""
) -> str:
    """State machine part implementing the 'compare-op' rule.

    compare-op ::=
        "smaller" "than" | "greater" "than" | "not" "equal" "to" |
        "equal" "to" | "at" "least" | "at" "most" | "approximately"
    """
    assert len(end_1locs) == 1

    locs = {"start": start_loc, "end": end_1locs[0], "tagname": tagname}
    locs.update(make_loc_names(prefix, "cmp", 5))

    text = """\
        cmp1; cmp2; cmp3; cmp4; cmp5;

        {start} -> cmp1 [SMALLER_KW] tag={tagname};
        {start} -> cmp1 [GREATER_KW] tag={tagname};
            cmp1 -> {end} [THAN_KW];
        {start} -> cmp2 [EQUAL_KW] tag={tagname};
            cmp2 -> {end} [TO_KW];
        {start} -> cmp3 [NOT_KW] tag={tagname};
            cmp3 -> cmp4 [EQUAL_KW];
            cmp4 -> {end} [TO_KW];
        {start} -> cmp5 [AT_KW];
            cmp5 -> {end} [LEAST_KW] tag={tagname};
            cmp5 -> {end} [MOST_KW] tag={tagname};
        {start} -> {end} [APPROXIMATELY_KW] tag={tagname};
    """

    return text.format(**locs)


def get_disjunctive_comparison_part(start_loc: str, end_2locs: List[str], prefix: str = "") -> str:
    """State machine part implementing the 'comparison-rule-line' comparisons.

    comparison-rule-line ::=
        comparison { "or" comparison }

    comparison ::=
        argument-name ( constraint-rule-literal | requirement-rule-literal )

    constraint-rule-literal ::=
        "is" compare-op bound

    requirement-rule-literal ::=
        auxiliary-verb "be" ( compare-op bound | objective )

    compare-op ::=
        "smaller" "than" | "greater" "than" | "not" "equal" "to" |
        "equal" "to" | "at" "least" | "at" "most" | "approximately"

    bound ::=
        argument-name | VALUE [ UNIT ] | "t.b.d." [ UNIT ]

    objective ::=
        "maximized" | "minimized"
    """
    assert len(end_2locs) == 2

    locs = {"start": start_loc, "end1": end_2locs[0], "end2": end_2locs[1]}
    locs.update(make_loc_names(prefix, "dis", 3))

    text = (
        """\
        {dis1}; {dis2}; {dis3};

        {start} -> {dis1} [DOTTEDNAME] tag=first_var;
    """
        + get_is_auxiliary_verb_part(locs["dis1"], [locs["dis2"]], tagname="is_aux")
        + """\
        {dis2} -> {end2} [MAXIMIZED_KW] tag=objective;
        {dis2} -> {end2} [MINIMIZED_KW] tag=objective;
    """
        + get_compare_op_part(locs["dis2"], [locs["dis3"]], tagname="compare_op")
        + """\
        {dis3} -> {end1} [NONCOMMA] tag=varvalue;
        {end1} -> {end2} [NONCOMMA] tag=unit;

        {end1} -> {start} [OR_KW] tag=or;
        {end2} -> {start} [OR_KW] tag=or;
    """
    )

    return text.format(**locs)



"""Module to export as an Excel workbook."""
from pathlib import Path
from typing import Dict, Optional, Union

from openpyxl.workbook import Workbook
from raesl.excel import defaults, sheets
from raesl.utils import get_scoped_nodes

# Compile module is excluded during docs generation.
try:
    from raesl.compile import to_graph
except ImportError:
    pass


def convert(
    *paths: Union[str, Path],
    output: Optional[Union[str, Path]] = defaults.OUTPUT,
    scopes: Dict[str, Optional[int]] = defaults.SCOPES,
) -> Workbook:
    """Write (part of) an ESL specification to a Excel workbook.

    Arguments:
        paths: ESL specification paths.
        output: Optional Workbook output path (will be overwritten without warning).
        scopes: Dictionary of component paths to relative depths of subcomponents to
            include as scopes for the generated output. Defaults to the complete tree.

    Returns:
        Excel workbook instance.
    """
    graph = to_graph(*paths)

    # Create workbook, but delete default sheet.
    wb = Workbook()
    wb.remove(wb["Sheet"])

    # Derive components from scopes.
    components = [node for node in get_scoped_nodes(graph, scopes) if node.kind == "component"]
    if not components:
        raise ValueError(f"No components found in selected scopes ('{scopes}'). Please reconsider.")

    # Add sheets.
    _, components = sheets.add_components_sheet(wb, components)
    _, goals = sheets.add_goals_sheet(wb, graph, components)
    _, transformations = sheets.add_transformations_sheet(wb, graph, components)
    _, designs = sheets.add_designs_sheet(wb, graph, components)
    _, behaviors = sheets.add_behaviors_sheet(wb, graph, components)
    _, needs = sheets.add_needs_sheet(wb, graph, components)
    _, variables = sheets.add_variable_sheet(wb, graph, components)

    sheets.add_overview_sheet(
        wb, graph, components, goals, transformations, designs, behaviors, needs
    )

    # Protect all sheets.
    for sheet in wb.sheetnames:
        wb[sheet].protection = defaults.SHEETPROTECTION

    # Write output if path is given.
    if output:
        wb.save(str(output))
    return wb



"""ESL to Doc

A Python package to process specifications in the form of `.esl` files into regular
document formats such as `.pdf`.
"""

from pathlib import Path
from typing import Any, Dict, Optional, Union

from raesl import logger
from raesl.doc.doc import Doc
from raesl.doc.locales import register_default_locale
from raesl.utils import check_output_path

register_default_locale()

LANGUAGE = "en"
TITLE = "ESL specification"
OUTPUT = "./esl.pdf"
PROLOGUE = None
EPILOGUE = None
RICH = "tex"
VARTABLE = True
FORCE = False
DRY = False


def convert(
    *paths: Union[str, Path],
    output: Union[str, Path] = OUTPUT,
    language: str = LANGUAGE,
    title: str = TITLE,
    prologue: Optional[Union[str, Path]] = PROLOGUE,
    epilogue: Optional[Union[str, Path]] = EPILOGUE,
    var_table: bool = VARTABLE,
    rich: Optional[str] = RICH,
    rich_opts: Optional[Dict[str, Any]] = None,
    force: bool = FORCE,
    dry: bool = DRY,
    **metadata,
):
    """Convert ESL files and/or directories to a formatted document.

    Arguments:
        paths: Paths to resolve into ESL files. May be any number of files and
            directories to scan.
        output: Optional output file (i.e. Markdown, PDF, DOCX).
        language: Output document language.
        title: Output document title.
        prologue: Optional prologue document to include (Markdown).
        epilogue: Optional epilogue document to include (Markdown).
        var_table: Add table with all variables within appendix.
        rich: Format of rich output to use. One of "tex", "md" or "off".
        rich_opts: Extra options for selected rich output.
        force: Whether to overwrite the output file or raise an error if the file
            already exists.
        dry: Dry run. Skip creating an output document.
    """
    output = Path(output)
    prologue = Path(prologue) if prologue else None
    epilogue = Path(epilogue) if epilogue else None

    logger.debug("Creating Doc object...")
    doc = Doc(
        *paths,
        language=language,
        title=title,
        prologue=prologue,
        epilogue=epilogue,
        var_table=var_table,
        rich=rich,
        rich_opts=rich_opts,
        **metadata,
    )
    logger.debug("Created Doc object!")

    if output:
        output = check_output_path(output, force)
        if dry:
            logger.debug("Dry run. Skipped writing to file.")
            return None
        doc.save(output)



"""Ratio ESL ``raesl`` Command Line Interface (CLI)."""
import click
import click_log

from raesl import __version__, logger
from raesl.doc.cli import doc
from raesl.excel.cli import excel
from raesl.jupyter.cli import jupyter
from raesl.server.cli import serve

# Compile module is excluded during docs generation.
try:
    from raesl.compile.cli import compile
except ImportError:
    pass

click_log.basic_config(logger=logger)


@click.group("raesl")
@click_log.simple_verbosity_option(logger)
@click.pass_context
def cli(ctx: click.Context):
    """Elephant Specification Language support by Ratio."""
    pass


@cli.command()
def version():
    """Print Ratio ESL version and exit."""
    logger.info(f"Ratio ESL version: {__version__}.")


cli.add_command(compile)
cli.add_command(doc)
cli.add_command(excel)
cli.add_command(jupyter)
cli.add_command(serve)

if __name__ == "__main__":
    cli()



"""Classes for the instantiated component graph."""
from typing import TYPE_CHECKING, List

from raesl.compile import diagnostics

if TYPE_CHECKING:
    from raesl.compile.ast.components import VarParam


class InstNode:
    """Instance node that connects one or more elementary nodes that are connected
    through parameters.

    Arguments:
        name: Dotted variable name of the associated elementary type.
        variable: Variable that created the node.

    Attributes:
        number: Unique number for each instance nodes, mostly useful for debugging.
        params: Parameters connected to the variable through this node.
        comments: Comments from the connected variable and parameters.
    """

    next_num = 1000

    def __init__(self, name: str, variable: "VarParam"):
        self.number = InstNode.next_num
        InstNode.next_num = self.number + 1

        self.name = name
        self.variable = variable
        self.params: List["VarParam"] = []
        self.comments: List[str] = []

    def add_param(self, param: "VarParam"):
        assert not param.is_variable
        self.params.append(param)

    def add_comment(self, words: List[str]):
        self.comments.extend(words)

    def owners(self):
        return [param for param in self.params if param.is_property]

    def check_owner(self, diag_store: diagnostics.DiagnosticStore):
        owners = self.owners()
        if len(owners) > 1:
            # Multiple property owners
            diag_store.add(
                diagnostics.E400(
                    self.name,
                    location=self.variable.name_tok.get_location(),
                    owners=[owner.name_tok.get_location() for owner in owners],
                )
            )

    def get_comment(self):
        return [p.comments for p in self.params]

    def __repr__(self):
        return "InstNode[" + str(self.number) + "]"



"""Code for collecting and adding behavior sections to component definitions."""
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from raesl.compile import diagnostics
from raesl.compile.ast import components
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.expr_checker import ExprChecker

if TYPE_CHECKING:
    from raesl.compile.ast.components import ComponentDefinition
    from raesl.compile.ast.exprs import Comparison, Disjunction, RelationComparison
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class ParsedBehavior:
    """Temporary storage of a behavior. This allows catching multiple default cases
    (these cannot be expressed in the AST). Also, it allows checking for receiving a
    sane order of calls from the parser.
    """

    def __init__(self, name: "Token", kind: str, cases: List["ParsedCase"]):
        self.name = name
        self.kind = kind
        self.cases = cases


# A case in behavior. The 'when_tok' and 'then_tok' can be set once for each case.
WhensType = List[Tuple["Token", Union["Disjunction", "RelationComparison"]]]
ThensType = List[Tuple["Token", "Comparison"]]


class ParsedCase:
    """Temporary storage for storing a case in a behavior while collecting the cases.

    Arguments:
        name: Name of the case.
        when_tok: Position of the 'when' line.
        then_tok: Position of the 'then' line.
        whens: Collected conditions. Will be empty for the default behavior. None
            means the variable should not be accessed at all.
        thens: Collected results.
    """

    def __init__(
        self,
        name: "Token",
        when_tok: Optional["Token"],
        then_tok: Optional["Token"],
        whens: WhensType,
        thens: ThensType,
    ):
        self.name = name
        self.when_tok = when_tok
        self.then_tok = then_tok
        self.whens: Optional[WhensType] = whens
        self.thens = thens


class CompDefBehaviorBuilder:
    """Class for constructing and checking behavior functions.

    Arguments:
        comp_child_builders: Storage of child builders for a component definition.

    Attributes:
        behavior_kind: Last seen kind of behavior ('requirement' or 'constraint').
        expect_conds: Whether the builder should allow conditions to be received from
            the parser.
        expect_results: Whether the builder should allow results to be received from
            the parser.
        pbehaviors: Collected behaviors.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.comp_child_builders = comp_child_builders

        self.behavior_kind: Optional[str] = None
        self.expect_conds = False
        self.expect_results = False
        self.pbehaviors: List[ParsedBehavior] = []  # Using List[ParsedCase] for its 'case's.

    def new_behavior_header(self, kind_tok: "Token"):
        """A new 'behavior' section header was found."""
        if kind_tok.tok_type == "BEHAVIOR_REQUIREMENT_KW":
            self.behavior_kind = components.REQUIREMENT
        else:
            assert kind_tok.tok_type == "BEHAVIOR_CONSTRAINT_KW"
            self.behavior_kind = components.CONSTRAINT

        self.expect_conds = False
        self.expect_results = False

    def new_behavior_function(self, label_tok: "Token"):
        """A new behavior functionality was found.

        Arguments:
            label_tok: Name of the new behavior.
        """
        assert self.behavior_kind is not None
        parsed_beh = ParsedBehavior(label_tok, self.behavior_kind, [])
        self.pbehaviors.append(parsed_beh)

        self.expect_conds = False
        self.expect_results = False

    def behavior_case(self, case_label_tok: "Token"):
        """A new case of the last started behavior functionality was found.

        Arguments:
            case_label_tok: Name of the case.
        """
        parsed_case = ParsedCase(case_label_tok, None, None, [], [])
        self.pbehaviors[-1].cases.append(parsed_case)

        self.expect_conds = False
        self.expect_results = False

    def behavior_normal_when(self, when_tok: "Token"):
        """The start of a normal 'when' condition block was found.

        Arguments:
            when_tok: Position of the start of the new block.
        """
        assert self.pbehaviors[-1].cases[-1].when_tok is None
        self.pbehaviors[-1].cases[-1].when_tok = when_tok

        # No need to setup case[-1].whens, as behavior_case() already did that.
        self.expect_conds = True
        self.expect_results = False

    def behavior_default_when(self, when_tok: "Token"):
        """The start of a default condition block was found.

        Arguments:
            when_tok: Position of the start of the new block.
        """
        assert self.pbehaviors[-1].cases[-1].when_tok is None
        self.pbehaviors[-1].cases[-1].when_tok = when_tok

        # Ensure code will crash if you add 'whens'.
        self.pbehaviors[-1].cases[-1].whens = None

        # We don't expect more 'when', and we must see a 'then' first.
        self.expect_conds = False
        self.expect_results = False

    def behavior_when_condition(
        self, name_tok: "Token", condition: Union["Disjunction", "RelationComparison"]
    ):
        """A new condition was found, add it to the last 'when' block.

        Arguments:
            name_tok: Name of the condition.
            condition: Condition to add.
        """
        assert self.expect_conds
        assert self.pbehaviors[-1].cases[-1].whens is not None
        self.pbehaviors[-1].cases[-1].whens.append((name_tok, condition))

    def behavior_normal_then(self, then_tok: "Token"):
        """The start of a 'then' result block was found.

        Arguments:
            then_tok: Position of the start of the new block.
        """
        assert self.pbehaviors[-1].cases[-1].then_tok is None
        self.pbehaviors[-1].cases[-1].then_tok = then_tok

        # No need to setup case[-1].thens, as behavior_case() already did that.
        self.expect_conds = False
        self.expect_results = True

    def behavior_then_result(self, name_tok: "Token", result: "Comparison"):
        """A new result was found, add it to the last 'then' block.

        Arguments:
            name_tok: Name of the result.
            result: Result to add.
        """
        assert self.expect_results
        self.pbehaviors[-1].cases[-1].thens.append((name_tok, result))

    def finish_comp(self, comp_def: "ComponentDefinition", _spec: "Specification"):
        """Verify correctness of the collected behavior and store good behavior in
        :obj:`comp_def`.

        Arguments:
            comp_def: Surrounding component definition supplying variables and
                parameters. Checked designs should be added to it after checking.
            _spec: Specification being constructed, source for types and verbs.
        """
        vps = utils.construct_var_param_map(comp_def)
        expr_checker = ExprChecker(vps, self.diag_store)

        beh_funcs = []
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for pbeh in self.pbehaviors:
            # Store labels of functions for duplicate checking.
            elements_by_label[pbeh.name.tok_text].append(pbeh)

            beh_func = components.BehaviorFunction(pbeh.kind, pbeh.name)
            default_cases: List["Token"] = []  # Positions of default cases in this function.
            cases_ordered_by_label: Dict[str, List["Token"]] = defaultdict(list)
            for pcase in pbeh.cases:
                cases_ordered_by_label[pcase.name.tok_text].append(pcase.name)

                # Process conditions.
                if pcase.whens is None:
                    # Default case.
                    assert pcase.when_tok is not None
                    default_cases.append(pcase.when_tok)
                    conditions = None
                else:
                    # Normal case.
                    conditions = self._convert_conditions(pcase.whens, expr_checker)

                # Process results.
                results = self._convert_results(pcase.thens, expr_checker)

                # Add case to the function.
                if conditions is None:
                    # Bluntly assume this happens at most once. If not, 'default_cases'
                    # will detect it and give an error.
                    beh_func.default_results = results
                else:
                    beh_case = components.BehaviorCase(pcase.name, conditions, results)
                    beh_func.cases.append(beh_case)

            # Verify uniqueness of cases.
            for dup_cases in cases_ordered_by_label.values():
                if len(dup_cases) > 1:
                    # Duplicate case names.
                    self.diag_store.add(
                        diagnostics.E200(
                            dup_cases[0].tok_text,
                            "behavior case",
                            location=pbeh.name.get_location(),
                            dupes=[dupe.get_location() for dupe in dup_cases],
                        )
                    )

            # Check number of default cases.
            if len(default_cases) > 1:
                # Duplicate fallback cases.
                self.diag_store.add(
                    diagnostics.E200(
                        pbeh.name.tok_text,
                        "fallback case",
                        location=pbeh.name.get_location(),
                        dupes=[dupe.get_location() for dupe in default_cases],
                    )
                )

            beh_funcs.append(beh_func)

        comp_def.behaviors = beh_funcs

    def _convert_conditions(
        self, whens: WhensType, expr_checker: ExprChecker
    ) -> List[components.BehaviorCondition]:
        """Check the conditions of a case, and convert them to :obj:`BehaviorCondition`
        instances.
        """
        conditions = []
        when_ordered_by_label: Dict[str, List["Token"]] = defaultdict(list)
        for when in whens:
            name_tok, cond = when
            when_ordered_by_label[name_tok.tok_text].append(name_tok)

            if expr_checker.check_expr(cond):
                conditions.append(components.BehaviorCondition(name_tok, cond))

        for dupes in when_ordered_by_label.values():
            if len(dupes) > 1:
                first = dupes[0]
                locs = [dupe.get_location() for dupe in dupes]
                self.diag_store.add(
                    diagnostics.E200(
                        first.tok_text,
                        "behavior condition",
                        location=first.get_location(),
                        dupes=locs,
                    )
                )

        return conditions

    def _convert_results(
        self, thens: ThensType, expr_checker: ExprChecker
    ) -> List[components.BehaviorResult]:
        """Check the results of a case, and convert them to :obj:`BehaviorResult`
        instances.
        """
        results = []
        then_ordered_by_label: Dict[str, List["Token"]] = defaultdict(list)
        for then in thens:
            name_tok, result = then
            then_ordered_by_label[name_tok.tok_text].append(name_tok)

            if expr_checker.check_expr(result):
                results.append(components.BehaviorResult(name_tok, result))

        for dupes in then_ordered_by_label.values():
            if len(dupes) > 1:
                first = dupes[0]
                locs = [dupe.get_location() for dupe in dupes]
                self.diag_store.add(
                    diagnostics.E200(
                        first.tok_text,
                        "behavior result",
                        first.get_location(),
                        dupes=locs,
                    )
                )

        return results
